from .auth_flow import AuthenticationFlow, AuthenticationResult, generate_otp
from .cookie_helpers import process_cookie_for_storage
from .deduplication import generate_deduplication_key
from .helpers import (
    check_login_opts,
    encode_chat_id,
    encode_digit,
    format_datetime,
    format_or_none,
    get_empty_vars,
    get_formatted_attr_or_load,
    get_utc,
    join_ids,
    parse_datetime,
    parse_iso_datetime,
    unescape_dict,
)
from .http_client import (
    HTTPClientConfig,
    RateLimitHandler,
    RequestResult,
    build_url,
    prepare_headers,
    should_retry_request,
)
from .lazy_loading import (
    LazyLoadBatch,
    LazyLoadingConfig,
    generate_cache_key,
    get_retry_delay,
)
from .ranges import combine_ranges

__all__ = (
    'AuthenticationFlow',
    'AuthenticationResult',
    'HTTPClientConfig',
    'LazyLoadBatch',
    'LazyLoadingConfig',
    'RateLimitHandler',
    'RequestResult',
    'build_url',
    'check_login_opts',
    'combine_ranges',
    'encode_chat_id',
    'encode_digit',
    'format_datetime',
    'format_or_none',
    'generate_cache_key',
    'generate_deduplication_key',
    'generate_otp',
    'get_empty_vars',
    'get_formatted_attr_or_load',
    'get_retry_delay',
    'get_utc',
    'join_ids',
    'parse_datetime',
    'parse_iso_datetime',
    'prepare_headers',
    'process_cookie_for_storage',
    'should_retry_request',
    'unescape_dict',
)



"""Constants for pararamio package."""

from .base import (
    BASE_API_URL,
    DATETIME_FORMAT,
    FILE_UPLOAD_URL,
    POSTS_LIMIT,
    REQUEST_TIMEOUT,
    UPLOAD_TIMEOUT,
    VERSION,
    XSRF_HEADER_NAME,
)

__all__ = [
    'BASE_API_URL',
    'DATETIME_FORMAT',
    'FILE_UPLOAD_URL',
    'POSTS_LIMIT',
    'REQUEST_TIMEOUT',
    'UPLOAD_TIMEOUT',
    'VERSION',
    'XSRF_HEADER_NAME',
]



"""Base model class for all models."""

from __future__ import annotations

from collections.abc import Mapping
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generic,
    Protocol,
    Self,
    TypeVar,
    cast,
)

if TYPE_CHECKING:
    from pararamio._core._types import FormatterT

__all__ = ('ClientT', 'CoreBaseModel', 'SerializationMixin')


class DictProtocol(Protocol):
    """Protocol for dict-like objects."""

    def __getitem__(self, key: str) -> Any: ...
    def __setitem__(self, key: str, value: Any) -> None: ...
    def get(self, key: str, default: Any = None) -> Any: ...


DataT = TypeVar('DataT', bound=Mapping[str, Any])
ClientT = TypeVar('ClientT')


class CoreBaseModel(Generic[DataT]):
    """Base class for all core models with shared data management."""

    _data: DataT
    _attr_formatters: ClassVar[FormatterT]

    def __repr__(self) -> str:
        """String representation of the model."""
        model_name = self.__class__.__name__
        id_value = getattr(self, 'id', None)
        return f'<{model_name}(id={id_value})>'


class AttrFormatterMixin(Generic[DataT]):
    """Mixin class that provides attribute formatting methods."""

    _data: DataT  # Type annotation for the data attribute
    _attr_formatters: ClassVar[FormatterT]

    def _get_formatted_attr(self, key: str) -> Any:
        if formatter := self._attr_formatters.get(key, None):
            return formatter(self._data, key)
        return self._data[key]


class SerializationMixin(Generic[ClientT, DataT]):
    """Mixin class that provides serialization and deserialization methods."""

    _data: DataT  # Type annotation for the data attribute

    def __init__(self, client: ClientT, **kwargs: Any) -> None:
        """Initialize with a client.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            **kwargs: Model data passed to parent classes
        """

    @classmethod
    def from_dict(
        cls,
        client: ClientT,
        data: DataT,
    ) -> Self:
        """Create model instance from dict data.

        Args:
            client: Client instance
            data: Raw API response data

        Returns:
            Model instance
        """
        return cls(client, **data)

    def to_dict(self) -> dict[str, Any]:
        """Convert model to dictionary.

        Returns:
            Model data as dict
        """
        result = {}
        for key, value in cast('dict[str, Any]', self._data).items():
            if isinstance(value, datetime):
                result[key] = value.isoformat()
            else:
                result[key] = value
        return result



"""Request deduplication utilities shared between sync and async implementations."""

from __future__ import annotations

import hashlib
from typing import Any

__all__ = ['generate_deduplication_key']


def generate_deduplication_key(method: str, *args: Any, **kwargs: Any) -> str:
    """Generate a deduplication key from method name and arguments.

    Args:
        method: Method name (e.g., 'get_user', 'get_chat')
        *args: Positional arguments
        **kwargs: Keyword arguments

    Returns:
        Unique key string

    Example:
        >>> key = generate_deduplication_key('get_user', 123)
        >>> # Result: 'get_user:123:hash_of_kwargs'
    """
    # Create base key from method and positional args
    parts = [method] + [str(arg) for arg in args]

    # Add sorted kwargs to ensure consistent ordering
    if kwargs:
        kwargs_str = ','.join(f'{k}={v}' for k, v in sorted(kwargs.items()))
        kwargs_hash = hashlib.md5(kwargs_str.encode()).hexdigest()[:8]
        parts.append(kwargs_hash)

    return ':'.join(parts)



"""Base API response schemas for generic data responses."""

from __future__ import annotations

from typing import Generic, TypedDict, TypeVar

# TypeVar for generic data content
T = TypeVar('T')


class DataResponse(TypedDict, Generic[T]):
    """Generic wrapper for API responses with 'data' field."""

    data: T


class DataListResponse(TypedDict, Generic[T]):
    """Generic wrapper for API responses with 'data' field containing a list."""

    data: list[T]


__all__ = ['DataListResponse', 'DataResponse', 'T']



"""Request deduplication utilities to avoid duplicate API calls."""

from __future__ import annotations

import logging
import threading
from collections.abc import Callable
from typing import Any, TypeVar

from pararamio._core.utils.deduplication import generate_deduplication_key

__all__ = ['RequestDeduplicator', 'generate_deduplication_key']

T = TypeVar('T')
log = logging.getLogger('pararamio.request_dedup')


class RequestDeduplicator:
    """Deduplicates identical in-flight requests to avoid unnecessary API calls.

    When multiple requests are made for the same resource simultaneously,
    only the first request is executed. Subsequent requests wait for the
    first request to complete and receive the same result.

    Thread-safe for concurrent requests.

    Examples:
        >>> dedup = RequestDeduplicator()
        >>> # Two simultaneous requests for the same user
        >>> def load_user(user_id):
        ...     return api.get_user(user_id)
        >>> # Only one API call is made, both get the same result
        >>> result1 = dedup.deduplicate('user:123', lambda: load_user(123))
        >>> result2 = dedup.deduplicate('user:123', lambda: load_user(123))
    """

    def __init__(self) -> None:
        """Initialize request deduplicator."""
        self._pending: dict[str, Any] = {}
        self._lock = threading.Lock()
        self._events: dict[str, threading.Event] = {}
        self._results: dict[str, Any] = {}

    def deduplicate(self, key: str, request_fn: Callable[[], T]) -> T:
        """Execute request with deduplication.

        If the same key is already in-flight, waits for that request to complete
        and returns its result instead of making a new request.

        Args:
            key: Unique key identifying the request (e.g., 'user:123', 'chat:456')
            request_fn: Function to execute if no duplicate request is in flight

        Returns:
            Result from the request function

        Example:
            >>> dedup = RequestDeduplicator()
            >>> result = dedup.deduplicate('user:123', lambda: client.get_user(123))
        """
        with self._lock:
            # Check if request is already in flight
            if key in self._pending:
                log.debug('Duplicate request detected: %s, waiting for completion', key)
                event = self._events[key]

        # If duplicate, wait for original request to complete
        if key in self._pending:
            event.wait()
            result = self._results.pop(key, None)
            log.debug('Duplicate request completed: %s', key)
            return result  # type: ignore[no-any-return]

        # Mark request as in-flight
        with self._lock:
            self._pending[key] = True
            event = threading.Event()
            self._events[key] = event

        try:
            # Execute the request
            log.debug('Executing request: %s', key)
            result = request_fn()

            # Store result for waiting requests
            with self._lock:
                self._results[key] = result

            # Notify waiting requests
            event.set()

            return result
        finally:
            # Cleanup
            with self._lock:
                self._pending.pop(key, None)
                self._events.pop(key, None)

    def generate_key(self, method: str, *args: Any, **kwargs: Any) -> str:
        """Generate a deduplication key from method name and arguments.

        Args:
            method: Method name (e.g., 'get_user', 'get_chat')
            *args: Positional arguments
            **kwargs: Keyword arguments

        Returns:
            Unique key string

        Example:
            >>> dedup = RequestDeduplicator()
            >>> key = dedup.generate_key('get_user', 123)
            >>> # Result: 'get_user:123:hash_of_kwargs'
        """
        return generate_deduplication_key(method, *args, **kwargs)

    def clear(self) -> None:
        """Clear all pending requests and cached results.

        Useful for testing or when you want to force fresh requests.
        """
        with self._lock:
            self._pending.clear()
            self._events.clear()
            self._results.clear()
        log.debug('Request deduplicator cleared')



"""Core File model without lazy loading."""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Unpack

from pararamio._core.api_schemas.responses import FileResponse
from pararamio._core.utils.helpers import parse_iso_datetime

from .base import ClientT, CoreBaseModel

if TYPE_CHECKING:
    from datetime import datetime

    from pararamio._core._types import FormatterT

__all__ = ('CoreFile',)


# Attribute formatters for File
FILE_ATTR_FORMATTERS: FormatterT = {
    'time_created': parse_iso_datetime,
}


class CoreFile(CoreBaseModel[FileResponse]):
    """Core File model with common functionality."""

    _data: FileResponse
    # File attributes
    guid: str
    name: str
    filename: str
    size: int
    mime_type: str
    origin: tuple[int, int] | None
    url: str | None
    path: str | None
    chat_id: int | None
    post_no: int | None
    user_id: int | None
    time_created: datetime | None

    _attr_formatters: ClassVar[FormatterT] = FILE_ATTR_FORMATTERS

    def __init__(
        self,
        client: ClientT,  # noqa: ARG002
        **kwargs: Unpack[FileResponse],
    ) -> None:
        """Initialize the file model with data.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            **kwargs: File data
        """
        # Ensure backward compatibility for 'name' -> 'filename'
        if 'name' in kwargs and 'filename' not in kwargs:
            kwargs['filename'] = kwargs['name']
        self._data = kwargs

    def __getattr__(self, name: str) -> object:
        """Get attribute from data dict.

        Args:
            name: Attribute name

        Returns:
            Attribute value

        Raises:
            AttributeError: If attribute not found
        """
        try:
            return self._data.get(name)
        except KeyError as e:
            raise AttributeError(
                f"'{self.__class__.__name__}' object has no attribute '{name}'"
            ) from e

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoreFile):
            return id(other) == id(self)
        return self.guid == other.guid



"""Team API response schemas."""

from __future__ import annotations

from typing import TypedDict


class TeamResponse(TypedDict, total=False):
    """Schema for Team API response."""

    # IDs
    id: int
    team_id: int  # Alias for id
    org_id: int  # Organization ID

    # Names
    title: str
    slug: str

    # Info
    description: str | None
    email_domain: str | None

    # Settings
    two_step_required: bool
    default_chat_id: int
    guest_chat_id: int | None
    state: str
    is_active: bool

    # Membership
    is_member: bool
    is_admin: bool
    inviter_id: int | None

    # Lists
    users: list[int]
    admins: list[int]
    groups: list[int]
    guests: list[int]

    # Timestamps (as ISO strings from API)
    time_created: str
    time_updated: str


class TeamMemberResponse(TypedDict, total=False):
    """Schema for Team Member API response."""

    # IDs
    id: int
    org_id: int

    # Info
    email: str
    phonenumber: str | None

    # Settings
    is_admin: bool
    is_member: bool
    two_step_enabled: bool
    inviter_id: int | None

    # Lists
    chats: list[int]
    groups: list[int]

    # State
    state: str
    last_activity: str | None

    # Timestamps (as ISO strings from API)
    time_created: str
    time_updated: str


class TeamMemberStatusResponse(TypedDict, total=False):
    """Schema for Team Member Status API response."""

    # IDs
    id: int
    user_id: int
    setter_id: int
    org_id: int

    # Info
    status: str

    # Timestamp (as ISO string from API)
    time_created: str


class TeamsResponse(TypedDict):
    """Response for GET /core/org with multiple teams."""

    orgs: list[TeamResponse]


class TeamMembersResponse(TypedDict):
    """Response for GET /core/org/{id}/member_info."""

    data: list[TeamMemberResponse]


class TeamStatusesResponse(TypedDict):
    """Response for GET /core/org/status."""

    data: list[TeamMemberStatusResponse]


class TeamSyncResponse(TypedDict):
    """Response for GET /core/org/sync."""

    ids: list[int]


__all__ = [
    'TeamMemberResponse',
    'TeamMemberStatusResponse',
    'TeamMembersResponse',
    'TeamResponse',
    'TeamStatusesResponse',
    'TeamSyncResponse',
    'TeamsResponse',
]



"""Cache helper utilities for sync operations."""

from __future__ import annotations

import logging
from collections.abc import Callable
from datetime import timedelta
from typing import TypeVar

# Import shared utilities from core
from pararamio._core.utils.cache_key import (
    deserialize_from_cache,
    generate_cache_key,
    serialize_for_cache,
)

from pararamio.protocols.cache import CacheProtocol

log = logging.getLogger('pararamio.cache.helpers')

T = TypeVar('T')


def cache_or_fetch(
    cache: CacheProtocol | None,
    key: str,
    fetch_func: Callable[[], T],
    ttl: timedelta | None = None,
) -> T:
    """Cache-or-fetch pattern helper.

    Tries to get value from cache first. If not found or cache is None,
    executes the fetch function and stores result in cache if available.

    Args:
        cache: Optional cache instance.
        key: Cache key.
        fetch_func: Function to call if cache miss.
        ttl: Optional TTL for cached value.

    Returns:
        The cached or fetched value.
    """
    # Try cache first if available
    if cache is not None:
        try:
            cached = cache.get(key)
            if cached is not None:
                log.debug('Cache hit for key: %s', key)
                return cached  # type: ignore[no-any-return]
        except Exception as e:
            log.warning('Cache get failed for key %s: %s', key, e)

    # Fetch from source
    log.debug('Cache miss for key: %s, fetching from source', key)
    result = fetch_func()

    # Store in cache if available
    if cache is not None and result is not None:
        try:
            cache.set(key, result, ttl=ttl)
            log.debug('Cached result for key: %s', key)
        except Exception as e:
            log.warning('Cache set failed for key %s: %s', key, e)

    return result


def invalidate_pattern(
    cache: CacheProtocol | None,
    pattern: str,
) -> int:
    """Invalidate cache entries matching a pattern.

    Args:
        cache: Optional cache instance.
        pattern: Key pattern/prefix to invalidate.

    Returns:
        Number of keys invalidated.
    """
    if cache is None:
        return 0

    try:
        count = cache.clear(prefix=pattern)
        log.debug('Invalidated %s cache entries with pattern: %s', count, pattern)
        return count
    except Exception as e:
        log.warning('Cache invalidation failed for pattern %s: %s', pattern, e)
        return 0


# Re-export from core for backward compatibility
__all__ = [
    'cache_or_fetch',
    'deserialize_from_cache',
    'generate_cache_key',
    'invalidate_pattern',
    'serialize_for_cache',
]



"""Models for pararamio package."""

from pararamio._core.api_schemas.responses.team import TeamMemberStatusResponse
from pararamio._core.models import CoreFile as File

from .activity import Activity, ActivityAction
from .attachment import Attachment
from .bot import PararamioBot
from .chat import Chat
from .deferred_post import DeferredPost
from .group import Group
from .poll import Poll, PollOption
from .post import Post
from .team import Team, TeamMember
from .user import User, UserSearchResult

__all__ = [
    'Activity',
    'ActivityAction',
    'Attachment',
    'Chat',
    'DeferredPost',
    'File',
    'Group',
    'PararamioBot',
    'Poll',
    'PollOption',
    'Post',
    'Team',
    'TeamMember',
    'TeamMemberStatusResponse',
    'User',
    'UserSearchResult',
]



"""API schemas for requests and responses."""

from .requests import (
    ChatCreateRequest,
    ChatUpdateSettingsRequest,
    DeferredPostCreateRequest,
    GroupEditRequest,
    MarkReadRequest,
    PostCreateRequest,
)
from .responses import (
    ActivityResponse,
    AttachmentResponse,
    ChatIdResponse,
    ChatListResponse,
    ChatResponseItem,
    ChatSearchResponse,
    ChatsResponse,
    ChatStatusItem,
    ChatSyncResponse,
    DataListResponse,
    DataResponse,
    DeferredPostCreateResponse,
    DeferredPostDeleteResponse,
    DeferredPostResponse,
    DeferredPostsResponse,
    DeleteFileResponse,
    EmptyResponse,
    FileResponse,
    GenericResponse,
    GroupOperationResponse,
    GroupResponseItem,
    GroupsResponse,
    KeywordsResponse,
    Mention,
    OkResponse,
    PollResponse,
    PostCreateResponse,
    PostDeleteResponse,
    PostEditResponse,
    PostEvent,
    PostResponseItem,
    PostsResponse,
    ReadStatusResponse,
    RerereResponse,
    TeamMemberResponse,
    TeamMembersResponse,
    TeamMemberStatusResponse,
    TeamResponse,
    TeamsResponse,
    TeamStatusesResponse,
    TeamSyncResponse,
    UserActivityResponse,
    UserLink,
    UserPrivateMessageResponse,
    UserResponse,
    UserResponseItem,
    UserSearchResponse,
    UsersResponse,
    WhoReadResponse,
)

__all__ = [
    'ActivityResponse',
    'AttachmentResponse',
    'ChatCreateRequest',
    'ChatIdResponse',
    'ChatListResponse',
    'ChatResponseItem',
    'ChatSearchResponse',
    'ChatStatusItem',
    'ChatSyncResponse',
    'ChatUpdateSettingsRequest',
    'ChatsResponse',
    'DataListResponse',
    'DataResponse',
    'DeferredPostCreateRequest',
    'DeferredPostCreateResponse',
    'DeferredPostDeleteResponse',
    'DeferredPostResponse',
    'DeferredPostsResponse',
    'DeleteFileResponse',
    'EmptyResponse',
    'FileResponse',
    'GenericResponse',
    'GroupEditRequest',
    'GroupOperationResponse',
    'GroupResponseItem',
    'GroupsResponse',
    'KeywordsResponse',
    'MarkReadRequest',
    'Mention',
    'OkResponse',
    'PollResponse',
    'PostCreateRequest',
    'PostCreateResponse',
    'PostDeleteResponse',
    'PostEditResponse',
    'PostEvent',
    'PostResponseItem',
    'PostsResponse',
    'ReadStatusResponse',
    'RerereResponse',
    'TeamMemberResponse',
    'TeamMemberStatusResponse',
    'TeamMembersResponse',
    'TeamResponse',
    'TeamStatusesResponse',
    'TeamSyncResponse',
    'TeamsResponse',
    'UserActivityResponse',
    'UserLink',
    'UserPrivateMessageResponse',
    'UserResponse',
    'UserResponseItem',
    'UserSearchResponse',
    'UsersResponse',
    'WhoReadResponse',
]



from pararamio._core._types import (
    BotProfileT,
    PostMetaFileT,
    PostMetaUserT,
    ProfileTypeT,
    QuoteRangeT,
    TextParsedT,
)

from .cache import InMemoryCache
from .client import Pararamio
from .constants import VERSION
from .cookie_manager import (
    CookieManager,
    FileCookieManager,
    InMemoryCookieManager,
    RedisCookieManager,
)
from .exceptions import (
    PararamioAuthenticationError,
    PararamioCaptchaAuthenticationError,
    PararamioException,
    PararamioHTTPRequestError,
    PararamioLimitExceededError,
    PararamioMethodNotAllowedError,
    PararamioPasswordAuthenticationError,
    PararamioRequestError,
    PararamioSecondFactorAuthenticationError,
    PararamioServerResponseError,
    PararamioValidationError,
    PararamioXSRFRequestError,
    PararamModelNotLoadedError,
    PararamMultipleFoundError,
    PararamNoNextPostError,
    PararamNoPrevPostError,
    PararamNotFoundError,
)
from .models import (
    Activity,
    ActivityAction,
    Attachment,
    Chat,
    DeferredPost,
    File,
    Group,
    PararamioBot,
    Poll,
    PollOption,
    Post,
    Team,
    TeamMember,
    User,
    UserSearchResult,
)
from .protocols.cache import CacheProtocol, CacheStatsProtocol

__version__ = VERSION


__all__ = (
    'Activity',
    'ActivityAction',
    'Attachment',
    'BotProfileT',
    'CacheProtocol',
    'CacheStatsProtocol',
    'Chat',
    'CookieManager',
    'DeferredPost',
    'File',
    'FileCookieManager',
    'Group',
    'InMemoryCache',
    'InMemoryCookieManager',
    'PararamModelNotLoadedError',
    'PararamMultipleFoundError',
    'PararamNoNextPostError',
    'PararamNoPrevPostError',
    'PararamNotFoundError',
    'Pararamio',
    'PararamioAuthenticationError',
    'PararamioBot',
    'PararamioCaptchaAuthenticationError',
    'PararamioException',
    'PararamioHTTPRequestError',
    'PararamioLimitExceededError',
    'PararamioMethodNotAllowedError',
    'PararamioPasswordAuthenticationError',
    'PararamioRequestError',
    'PararamioSecondFactorAuthenticationError',
    'PararamioServerResponseError',
    'PararamioValidationError',
    'PararamioXSRFRequestError',
    'Poll',
    'PollOption',
    'Post',
    'PostMetaFileT',
    'PostMetaUserT',
    'ProfileTypeT',
    'QuoteRangeT',
    'RedisCookieManager',
    'Team',
    'TeamMember',
    'TextParsedT',
    'User',
    'UserSearchResult',
)



"""Core Activity model without lazy loading."""

from __future__ import annotations

import datetime as dt
from enum import Enum
from typing import TYPE_CHECKING, Any, ClassVar, Unpack

from pararamio._core.api_schemas.responses import ActivityResponse
from pararamio._core.utils.helpers import parse_iso_datetime

from .base import CoreBaseModel

if TYPE_CHECKING:
    from collections.abc import Callable

    from pararamio._core._types import FormatterT

__all__ = ('Activity', 'ActivityAction', 'CoreActivity')


# Attribute formatters for Activity
ACTIVITY_ATTR_FORMATTERS: FormatterT = {
    'datetime': parse_iso_datetime,
}


class ActivityAction(Enum):
    """Activity action types."""

    ONLINE = 'online'
    OFFLINE = 'offline'
    AWAY = 'away'
    READ = 'thread-read'
    POST = 'thread-post'
    CALL = 'calling'
    CALL_END = 'endcall'


class Activity:
    """User activity record."""

    action: ActivityAction
    time: dt.datetime

    def __init__(self, action: ActivityAction, time: dt.datetime) -> None:
        """Initialize activity.

        Args:
            action: Activity action type
            time: Activity timestamp
        """
        self.action = action
        self.time = time

    def __str__(self) -> str:
        """String representation."""
        return f'Activity({self.time}, {self.action.value})'

    def __repr__(self) -> str:
        """Detailed representation."""
        return f'<Activity(action={self.action}, time={self.time})>'

    def __eq__(self, other: object) -> bool:
        """Check equality."""
        if not isinstance(other, Activity):
            return False
        return self.action == other.action and self.time == other.time

    @classmethod
    def from_api_data(cls, data: dict[str, str]) -> Activity:
        """Create Activity from API response data.

        Args:
            data: API response data

        Returns:
            Activity instance

        Raises:
            ValueError: If the time format is invalid
        """
        time = parse_iso_datetime(data, 'datetime')
        if time is None:
            raise ValueError('Invalid time format')

        return cls(
            action=ActivityAction(data['action']),
            time=time,
        )


class CoreActivity(CoreBaseModel[ActivityResponse]):
    """Core Activity model with common functionality."""

    _data: ActivityResponse
    # Activity attributes
    action: str
    datetime: dt.datetime

    _attr_formatters: ClassVar[FormatterT] = ACTIVITY_ATTR_FORMATTERS

    def __init__(self, client: Any, **kwargs: Unpack[ActivityResponse]) -> None:  # noqa: ARG002
        """Initialize the activity model with data.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            **kwargs: Activity data
        """
        self._data = kwargs

    def __str__(self) -> str:
        return self.action

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoreActivity):
            return id(other) == id(self)
        return self.action == other.action and self.datetime == other.datetime

    @staticmethod
    def filter_activities_by_date_range(
        data: list[dict[str, Any]],
        start: dt.datetime,
        end: dt.datetime,
        activity_factory: Callable[[dict[str, Any]], Any],
    ) -> tuple[list[Any], bool]:
        """Filter activities by date range and create activity objects.

        Args:
            data: List of raw activity data from API
            start: Start datetime for filtering
            end: End datetime for filtering
            activity_factory: Factory function to create activity instances

        Returns:
            Tuple of (the filtered activities list, is_last_page flag)
        """
        results = []
        is_last_page = False

        for activity_data in data:
            activity = activity_factory(activity_data)

            # Skip activities after end date
            if hasattr(activity, 'time'):
                activity_time = activity.time
            elif hasattr(activity, 'datetime'):
                activity_time = activity.datetime
            else:
                continue

            if activity_time > end:
                continue

            # Stop processing if we've reached activities before start date
            if activity_time < start:
                is_last_page = True
                break

            results.append(activity)

        return results, is_last_page



"""Core Bot model without lazy loading."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar

if TYPE_CHECKING:
    from pararamio._core._types import FormatterT

__all__ = ('CoreBot',)


# Attribute formatters for Bot
BOT_ATTR_FORMATTERS: FormatterT = {}


class CoreBot:
    """Core Bot model with common functionality.

    Note: This is a minimal bot model for data representation.
    The actual bot client with request methods is in AsyncPararamioBot/PararamioBot.
    """

    # Bot attributes
    _key: str
    name: str | None
    id: int | None
    chat_id: int | None
    organization_id: int | None

    _attr_formatters: ClassVar[FormatterT] = BOT_ATTR_FORMATTERS

    def __init__(self, key: str) -> None:
        """Initialize a bot model with data.

        Args:
            key: Bot API key
        """
        if len(key) > 50:
            key = key[20:]
        self._key = key

    @staticmethod
    def prepare_post_message_data(
        chat_id: int, text: str, reply_no: int | None = None
    ) -> dict[str, Any]:
        """Prepare data for post_message API call.

        Args:
            chat_id: Target chat ID
            text: Message text
            reply_no: Optional message number to reply to

        Returns:
            Dictionary with message data for API call
        """
        data: dict[str, Any] = {
            'chat_id': chat_id,
            'text': text,
        }
        if reply_no is not None:
            data['reply_no'] = reply_no
        return data



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Unpack

from pararamio._core.api_schemas.requests import DeferredPostCreateRequest
from pararamio._core.api_schemas.responses import (
    DeferredPostCreateResponse,
    DeferredPostDeleteResponse,
    DeferredPostResponse,
    DeferredPostsResponse,
)
from pararamio._core.models.base import SerializationMixin
from pararamio._core.models.deferred_post import CoreDeferredPost
from pararamio._core.utils.helpers import format_datetime

# Imports from core
from pararamio.exceptions import PararamNotFoundError

from .base import SyncClientMixin

if TYPE_CHECKING:
    from pararamio.client import Pararamio

__all__ = ('DeferredPost',)


class DeferredPost(
    CoreDeferredPost['Pararamio'],
    SyncClientMixin[DeferredPostResponse],
    SerializationMixin['Pararamio', DeferredPostResponse],
):
    """Sync DeferredPost model for scheduled posts."""

    def __init__(self, client: Pararamio, **kwargs: Unpack[DeferredPostResponse]) -> None:
        """Initialize sync deferred post.

        Args:
            client: Pararamio client
            **kwargs: Additional post data including id
        """
        # Ensure id is set
        if 'id' not in kwargs:
            kwargs['id'] = 0

        # Use super() to initialize both parent classes properly
        super().__init__(client, **kwargs)
        # Mark as loaded if we have the required fields
        if 'id' in kwargs:
            self._set_loaded()

    def __str__(self) -> str:
        text = self._data.get('text', None)
        if text is None:
            self.load()
            text = self._data['text']
        return text

    def load(self) -> DeferredPost:
        for post in self.get_deferred_posts(self._client):
            if post.id == self.id:
                self._data = post._data
                return self
        raise PararamNotFoundError(f'Deferred post with id {self.id} not found')

    def delete(self) -> bool:
        """Delete this deferred post.

        Returns:
            True if successful
        """
        url = f'/msg/deferred/{self.id}'
        response = self._client.api_delete(url, response_model=DeferredPostDeleteResponse)
        return response.get('result') == 'OK'

    @classmethod
    def create(
        cls,
        client: Pararamio,
        chat_id: int,
        text: str,
        *,
        time_sending: datetime,
        reply_no: int | None = None,
        quote_range: tuple[int, int] | None = None,
    ) -> DeferredPost:
        """Create a new deferred (scheduled) post.

        Args:
            client: Pararamio client
            chat_id: Target chat ID
            text: Post text
            time_sending: When to send the post
            reply_no: Optional post number to reply to
            quote_range: Optional quote range as (start, end) tuple

        Returns:
            Created DeferredPost object
        """
        url = '/msg/deferred'

        # Use TypedDict for type-safe data construction
        data: DeferredPostCreateRequest = {
            'chat_id': chat_id,
            'text': text,
            'time_sending': format_datetime(time_sending),
        }

        # Add optional fields only if provided
        if reply_no is not None:
            data['reply_no'] = reply_no
        if quote_range is not None:
            data['quote_range'] = quote_range

        response = client.api_post(url, dict(data), response_model=DeferredPostCreateResponse)

        return cls(
            client,
            id=int(response['deferred_post_id']),
            **data,
        )

    @classmethod
    def get_deferred_posts(cls, client: Pararamio) -> list[DeferredPost]:
        """Get all deferred posts for the current user.

        Args:
            client: Pararamio client

        Returns:
            List of DeferredPost objects
        """
        url = '/msg/deferred'
        response = client.api_get(url, response_model=DeferredPostsResponse)
        posts_data = response.get('posts', [])

        return [cls(client, **post_data) for post_data in posts_data]



"""Cookie management base mixin shared between sync and async implementations."""

from __future__ import annotations

import json
import logging
from datetime import UTC, datetime
from http.cookiejar import Cookie, CookieJar
from typing import Any

log = logging.getLogger(__name__)


class CookieManagerBaseMixin:
    """Mixin with common cookie management functionality."""

    # These attributes are expected to be defined by classes using this mixin
    _cookies: dict[str, Cookie]
    _version: int

    @staticmethod
    def make_key(cookie: Cookie) -> str:
        """Make a unique key for cookie."""
        return f'{cookie.domain}:{cookie.path}:{cookie.name}'

    @staticmethod
    def cookie_to_dict(cookie: Cookie) -> dict[str, Any]:
        """Convert Cookie object to dictionary."""
        return {
            'name': cookie.name,
            'value': cookie.value,
            'domain': cookie.domain,
            'path': cookie.path,
            'secure': cookie.secure,
            'expires': cookie.expires,
            'discard': cookie.discard,
            'comment': cookie.comment,
            'comment_url': cookie.comment_url,
            'rfc2109': cookie.rfc2109,
            'port': cookie.port,
            'port_specified': cookie.port_specified,
            'domain_specified': cookie.domain_specified,
            'domain_initial_dot': cookie.domain_initial_dot,
            'path_specified': cookie.path_specified,
            'version': cookie.version,
        }

    @staticmethod
    def dict_to_cookie(data: dict[str, Any]) -> Cookie | None:
        """Convert dictionary to a Cookie object."""
        try:
            return Cookie(
                version=data.get('version', 0),
                name=data['name'],
                value=data['value'],
                port=data.get('port'),
                port_specified=data.get('port_specified', False),
                domain=data['domain'],
                domain_specified=data.get('domain_specified', True),
                domain_initial_dot=data.get('domain_initial_dot', False),
                path=data['path'],
                path_specified=data.get('path_specified', True),
                secure=data.get('secure', False),
                expires=data.get('expires'),
                discard=data.get('discard', False),
                comment=data.get('comment'),
                comment_url=data.get('comment_url'),
                rest={},
                rfc2109=data.get('rfc2109', False),
            )
        except (KeyError, TypeError):
            log.exception('Failed to create cookie from dict')
            return None

    def populate_jar(self, cookie_jar: CookieJar) -> None:
        """Populate a CookieJar with our cookies."""
        cookies = getattr(self, '_cookies', {})
        # Check if we need thread safety
        lock = getattr(self, '_lock', None)
        if lock is not None:
            with lock:
                for cookie in cookies.values():
                    cookie_jar.set_cookie(cookie)
        else:
            for cookie in cookies.values():
                cookie_jar.set_cookie(cookie)

    def has_cookies(self) -> bool:
        """Check if manager has any cookies."""
        return bool(getattr(self, '_cookies', {}))

    def _get_file_version(self) -> int:
        """Get current version from the file."""
        version_path = getattr(self, 'version_path', None)
        if not version_path or not version_path.exists():
            return 0

        try:
            with version_path.open(encoding='utf-8') as f:
                return int(f.read().strip())
        except (OSError, ValueError):
            return 0

    def _load_cookies_from_json(self, data: str | None) -> bool:
        """Load cookies from JSON data.

        Args:
            data: JSON string containing cookie's data

        Returns:
            True if loaded successfully, False otherwise
        """
        if not data:
            return False

        parsed_data = json.loads(data)
        return self._load_cookies_from_dict(parsed_data)

    def _load_cookies_from_dict(self, data: dict[str, Any]) -> bool:
        """Load cookies from dictionary data.

        Args:
            data: Dictionary containing cookie's data

        Returns:
            True if loaded successfully
        """
        self._version = data.get('version', 0)
        cookies_data = data.get('cookies', [])

        self._cookies.clear()
        for cookie_dict in cookies_data:
            cookie = self.dict_to_cookie(cookie_dict)
            if cookie:
                key = self.make_key(cookie)
                self._cookies[key] = cookie

        return True

    def _prepare_cookies_data(self) -> dict[str, Any]:
        """Prepare cookie's data for saving.

        Returns:
            Dictionary with a version, cookies, and timestamp
        """
        cookies = getattr(self, '_cookies', {})
        version = getattr(self, '_version', 0)

        cookies_data = [
            self.cookie_to_dict(cookie)
            for cookie in cookies.values()
            if not cookie.discard  # Only save persistent cookies
        ]

        return {
            'version': version,
            'cookies': cookies_data,
            'saved_at': datetime.now(UTC).isoformat(),
        }

    def _increment_file_version(self) -> int:
        """Increment version and save to file."""
        version_path = getattr(self, 'version_path', None)
        if not version_path:
            return 0

        self._version = self._get_file_version() + 1

        try:
            version_path.parent.mkdir(parents=True, exist_ok=True)
            with version_path.open('w', encoding='utf-8') as f:
                f.write(str(self._version))
        except OSError:
            log.exception('Failed to save version')

        return self._version



"""Core Attachment model without lazy loading."""

from __future__ import annotations

import mimetypes
import os
from io import BufferedRandom, BufferedReader, BytesIO
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING, Any, BinaryIO, ClassVar, Unpack

from pararamio._core.api_schemas.responses import AttachmentResponse

from .base import CoreBaseModel

if TYPE_CHECKING:
    from pararamio._core._types import FormatterT

__all__ = ('CoreAttachment', 'guess_mime_type')


def guess_mime_type(filename: str | PathLike[str]) -> str:
    """Guess MIME type from filename.

    Args:
        filename: File name or path

    Returns:
        MIME type string
    """
    if not mimetypes.inited:
        mimetypes.init(files=os.environ.get('PARARAMIO_MIME_TYPES_PATH', None))
    return mimetypes.guess_type(str(filename))[0] or 'application/octet-stream'


# Attribute formatters for Attachment
ATTACHMENT_ATTR_FORMATTERS: FormatterT = {}


class CoreAttachment(CoreBaseModel[AttachmentResponse]):
    """Core Attachment model with common functionality."""

    _data: AttachmentResponse
    # Attachment attributes
    guid: str
    name: str
    size: int
    mime_type: str
    url: str
    post_no: int | None
    chat_id: int | None

    _attr_formatters: ClassVar[FormatterT] = ATTACHMENT_ATTR_FORMATTERS

    def __init__(
        self,
        client: Any | None = None,  # noqa: ARG002
        **kwargs: Unpack[AttachmentResponse],
    ) -> None:
        """Initialize attachment model with data.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            **kwargs: Attachment data
        """
        self._data = kwargs

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoreAttachment):
            return id(other) == id(self)
        return self.guid == other.guid

    @staticmethod
    def guess_filename_from_file(
        file: str | bytes | PathLike[str] | BytesIO | BinaryIO, filename: str | None = None
    ) -> str:
        """Guess filename from a file object.

        Args:
            file: File object (path, BytesIO, or BinaryIO)
            filename: Optional explicit filename

        Returns:
            Guessed filename or 'unknown'
        """
        if filename:
            return filename

        if isinstance(file, str | PathLike):
            return Path(file).name

        if isinstance(file, BytesIO | BinaryIO | BufferedReader | BufferedRandom):
            try:
                name = getattr(file, 'name', None)
                if name:
                    return Path(name).name
            except AttributeError:
                pass

        return 'unknown'

    @staticmethod
    def guess_content_type_from_file(
        file: str | bytes | PathLike[str] | BytesIO | BinaryIO | BufferedReader | BufferedRandom,
        content_type: str | None = None,
    ) -> str:
        """Guess content type from a file object.

        Args:
            file: File object
            content_type: Optional explicit content type

        Returns:
            MIME type string
        """
        if content_type:
            return content_type

        if isinstance(file, str | PathLike):
            return guess_mime_type(file)

        if isinstance(file, BytesIO | BinaryIO | BufferedReader | BufferedRandom):
            name = getattr(file, 'name', None)
            if name:
                return guess_mime_type(name)

        return 'application/octet-stream'

    @staticmethod
    def get_file_pointer(
        file: str | bytes | PathLike[str] | BytesIO | BinaryIO,
    ) -> BytesIO | BinaryIO:
        """Get file pointer from various file types.

        Args:
            file: File object (bytes, path, BytesIO, or BinaryIO)

        Returns:
            File-like object (BytesIO or BinaryIO)

        Raises:
            TypeError: If the file type is not supported
        """
        if isinstance(file, bytes):
            return BytesIO(file)

        if isinstance(file, str | PathLike):
            with Path(file).open('rb') as f:
                return BytesIO(f.read())

        if isinstance(file, BytesIO | BinaryIO | BufferedReader | BufferedRandom):
            return file

        raise TypeError(f'Unsupported type {type(file)}')



"""Core Post model without lazy loading."""

from __future__ import annotations

import os
from datetime import datetime
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar, Unpack, cast

from pararamio._core.api_schemas.responses import FileUploadFields, PostResponseItem
from pararamio._core.exceptions import (
    PararamioTypeError,
    PararamioValidationError,
    PararamioValueError,
)
from pararamio._core.utils.helpers import parse_iso_datetime

from .base import CoreBaseModel

if TYPE_CHECKING:
    from pararamio._core._types import FormatterT, PostMetaT, TextParsedT
    from pararamio._core.api_schemas import PostEvent
    from pararamio._core.api_schemas.responses import Mention, UserLink

    from .chat import CoreChat

__all__ = ('CorePost',)


# Attribute formatters for Post
POST_ATTR_FORMATTERS: FormatterT = {
    'post_no': lambda data, key: int(data[key]),
    'time_edited': parse_iso_datetime,
    'time_created': parse_iso_datetime,
}


ChatT = TypeVar('ChatT', bound='CoreChat[Any]')


class CorePost(CoreBaseModel[PostResponseItem], Generic[ChatT]):
    """Core Post model with common functionality."""

    # Post attributes from actual API response
    _chat: ChatT
    _data: PostResponseItem
    uuid: str
    id: int | None
    post_no: int
    user_id: int
    reply_no: int | None
    event: PostEvent | None
    meta: PostMetaT
    ver: int | None
    is_deleted: bool
    time_created: datetime
    time_edited: datetime | None
    text: str
    text_parsed: list[TextParsedT]
    _attr_formatters = POST_ATTR_FORMATTERS

    def __init__(  # type: ignore[misc]
        self,
        chat: ChatT,
        post_no: int | None = None,
        **kwargs: Unpack[PostResponseItem],
    ) -> None:
        self._chat = chat
        # Handle positional post_no
        if post_no is not None:
            if 'post_no' in kwargs:
                kwargs.pop('post_no')
            kwargs['post_no'] = post_no
        self._data = cast('PostResponseItem', kwargs)

        # Validate that required ID is present
        if 'post_no' not in self._data:
            raise PararamioValidationError(
                'Post requires post_no to be present in data. '
                'Cannot create a Post without a valid post number.'
            )

        super().__init__(chat._client, **kwargs)  # type: ignore[call-arg]

    @classmethod
    def from_dict(
        cls,
        chat: ChatT,
        data: PostResponseItem,
    ) -> Self:
        """Create Post instance from dict data.

        Args:
            chat: Parent chat object
            data: Post response data

        Returns:
            Post instance marked as loaded
        """
        # Note: CorePost doesn't have _set_loaded, but sync/async implementations do
        # This is called on the actual subclass (Post), not CorePost directly
        return cls(chat, **data)

    def to_dict(self) -> PostResponseItem:
        """Convert Post to dictionary.

        Returns:
            Post data as dict
        """
        return self._data.copy()

    @property
    def chat(self) -> ChatT:
        """Get chat object."""
        return self._chat

    @property
    def chat_id(self) -> int:
        """Get chat ID. Prefer from _data if present, otherwise from parent chat."""
        # Try to get from post data first
        if 'chat_id' in self._data:
            return self._data['chat_id']

        # Try chat._data directly to avoid triggering load (if it's a dict)
        if hasattr(self._chat, '_data') and isinstance(self._chat._data, dict):
            if 'chat_id' in self._chat._data:
                return self._chat._data['chat_id']
            if 'id' in self._chat._data:
                return self._chat._data['id']

        # Last resort - may trigger lazy loading or use mock attribute
        return self._chat.id

    @property
    def is_reply(self) -> bool:
        """Check if this post is a reply."""
        return self.reply_no is not None

    @property
    def is_event(self) -> bool:
        """Check if this is an event post."""
        return bool(self.event)

    @property
    def mentions(self) -> list[Mention]:
        """Get post-mentions."""
        return [
            {
                'id': item['id'],
                'name': item['name'],
                'value': item['value'],
            }
            for item in self.text_parsed
            if item.get('type') == 'mention'
        ]

    @property
    def is_bot(self) -> bool:
        """Check if post is from a bot."""
        return self.meta.get('user', {}).get('is_bot', False)

    @property
    def is_file(self) -> bool:
        """Check if post contains file attachment."""
        return 'file' in self.meta

    @property
    def is_mention(self) -> bool:
        """Check if post contains mentions.

        Returns:
            True if post has mentions
        """
        if self.text_parsed is None:
            return False
        return any(item.get('type') == 'mention' for item in self.text_parsed)

    @property
    def user_links(self) -> list[UserLink]:
        """Get user links in post."""
        return [
            {
                'id': item['id'],
                'name': item['name'],
                'value': item['value'],
            }
            for item in self.text_parsed
            if item.get('type') == 'user_link' or item.get('type') == 'userlink'
        ]

    @property
    def attachments(self) -> list[str]:
        """Get attachments in post."""
        return self.meta.get('attachments', [])

    def _compare_validations(self, other: object) -> None:
        """Validate that posts can be compared."""
        if not isinstance(other, CorePost):
            raise PararamioTypeError(f'can not compare post and {other.__class__.__name__}')
        if self.chat_id != other.chat_id:
            raise PararamioValueError('can not compare posts from different chats')

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CorePost):
            return id(other) == id(self)
        return self.chat_id == other.chat_id and self.post_no == other.post_no

    def __lt__(self, other: object) -> bool:
        self._compare_validations(other)
        assert isinstance(other, CorePost)  # for type checker
        return self.post_no < other.post_no

    def __le__(self, other: object) -> bool:
        self._compare_validations(other)
        assert isinstance(other, CorePost)  # for type checker
        return self.post_no <= other.post_no

    def __gt__(self, other: object) -> bool:
        self._compare_validations(other)
        assert isinstance(other, CorePost)  # for type checker
        return self.post_no > other.post_no

    def __ge__(self, other: object) -> bool:
        self._compare_validations(other)
        assert isinstance(other, CorePost)  # for type checker
        return self.post_no >= other.post_no

    def __str__(self) -> str:
        return self.text

    def __repr__(self) -> str:
        """Detailed representation."""
        return (
            f'<Post(client={hex(id(self.chat.client))}, chat_id={self.chat_id}, post_no'
            f'={self.post_no})>'
        )

    @staticmethod
    def prepare_file_upload_fields(
        file: Any,  # BinaryIO | BytesIO
        chat_id: int | None,
        *,
        filename: str | None = None,
        type_: str | None = None,
        organization_id: int | None = None,
        reply_no: int | None = None,
        quote_range: str | None = None,
    ) -> tuple[FileUploadFields, str | None]:
        """Prepare fields and validate parameters for file upload.

        Args:
            file: File-like object to upload
            chat_id: Chat ID for the upload
            filename: Optional filename
            type_: Optional file type (can be 'organization_avatar', 'chat_avatar', or content type)
            organization_id: Optional organization ID (required for organization_avatar)
            reply_no: Optional reply number
            quote_range: Optional quote range

        Returns:
            Tuple of (FileUploadFields dict, content_type)

        Raises:
            PararamioValidationError: If validation fails
        """
        # Validate required parameters
        if type_ is None and not filename:
            raise PararamioValidationError('filename must be set when type is None')

        if type_ == 'organization_avatar' and organization_id is None:
            raise PararamioValidationError(
                'organization_id must be set when type is organization_avatar'
            )

        if type_ == 'chat_avatar' and chat_id is None:
            raise PararamioValidationError('chat_id must be set when type is chat_avatar')

        # Determine content type
        content_type = None
        if type_ not in ('organization_avatar', 'chat_avatar'):
            content_type = type_

        # Get file size
        file.seek(0, os.SEEK_END)
        file_size = file.tell()
        file.seek(0, 0)

        # Prepare fields
        fields = FileUploadFields(
            type=type_,
            filename=filename,
            size=file_size,
            chat_id=chat_id,
            organization_id=organization_id,
            reply_no=reply_no,
            quote_range=quote_range,
        )

        return fields, content_type



from __future__ import annotations

from typing import TYPE_CHECKING, Unpack

from pararamio._core import PararamioRequestError
from pararamio._core.api_schemas import (
    OkResponse,
    TeamMemberResponse,
    TeamMembersResponse,
    TeamMemberStatusResponse,
    TeamResponse,
    TeamsResponse,
    TeamStatusesResponse,
    TeamSyncResponse,
)
from pararamio._core.models import (
    CoreTeam,
    CoreTeamMember,
    SerializationMixin,
)

from .base import SyncClientMixin
from .group import Group
from .user import User

if TYPE_CHECKING:
    from pararamio.client import Pararamio
__all__ = (
    'Team',
    'TeamMember',
)


class TeamMember(
    CoreTeamMember['Pararamio'],
    SyncClientMixin[TeamMemberResponse],
    SerializationMixin['Pararamio', TeamMemberResponse],
):
    """Sync TeamMember model."""

    def __init__(
        self,
        client: Pararamio,
        **kwargs: Unpack[TeamMemberResponse],
    ) -> None:
        """Initialize sync team member.

        Args:
            client: Pararamio client
            **kwargs: Team member data
        """
        super().__init__(client, **kwargs)

    @property
    def user(self) -> User:
        return User(client=self.client, id=self.id)

    def get_last_status(self) -> TeamMemberStatusResponse | None:
        url = f'/core/org/status?user_ids={self.id}'
        res = self.client.api_get(url, response_model=TeamStatusesResponse).get('data', [])
        if not res:
            return None
        return res[0]

    def add_status(self, status: str) -> bool:
        url = '/core/org/status'
        data = {
            'org_id': self.org_id,
            'status': status,
            'user_id': self.id,
        }
        res = self.client.api_post(url, data=data, response_model=OkResponse)
        return bool(res.get('result') == 'OK')


class Team(
    CoreTeam['Pararamio'],
    SyncClientMixin[TeamResponse],
    SerializationMixin['Pararamio', TeamResponse],
):
    """Sync Team model with lazy loading support."""

    def __init__(
        self,
        client: Pararamio,
        **kwargs: Unpack[TeamResponse],
    ) -> None:
        """Initialize sync team.

        Args:
            client: Pararamio client
            **kwargs: Team data
        """
        super().__init__(client, **kwargs)

    def create_role(self, name: str, description: str | None = None) -> Group:
        return Group.create(
            self.client, organization_id=self.id, name=name, description=description
        )

    def load(self) -> Team:
        """
        Fetches data from the API for the current organization's ID
        and updates the object's data with caching.

        Requests data from the organization's endpoint using the object's ID,
        then updates the object's data with the response.

        Returns:
            self: The current object instance with updated data.
        """
        # Try to load from cache first if available
        if self.client._cache:
            cache_key = f'team.{self.id}'
            cached = self.client._cache.get(cache_key)
            if cached:
                self._data.update(cached)
                return self

        # Load from API if not cached
        url = f'/core/org?ids={self.id}'
        res = self.client.api_get(url, response_model=TeamsResponse)
        if res and 'orgs' in res and res['orgs']:
            self._data.update(res['orgs'][0])

            # Cache the team data if cache is available
            if self.client._cache:
                cache_key = f'team.{self.id}'
                self.client._cache.set(cache_key, res['orgs'][0])

        return self

    def get_member_info(self, user_id: int) -> TeamMember:
        url = f'/core/org/{self.id}/member_info/{user_id}'
        res = self.client.api_get(url, response_model=TeamMemberResponse)
        if not res:
            raise PararamioRequestError(f'empty response for user {user_id}')
        return TeamMember(self.client, **res)

    def get_members_info(self) -> list[TeamMember]:
        url = f'/core/org/{self.id}/member_info'
        res = self.client.api_get(url, response_model=TeamMembersResponse)
        if res:
            return [TeamMember(self.client, **m) for m in res.get('data', [])]
        return []

    @classmethod
    def get_my_team_ids(cls, client: Pararamio) -> list[int]:
        url = '/core/org/sync'
        res = client.api_get(url, response_model=TeamSyncResponse)
        return res.get('ids', [])

    @classmethod
    def load_teams(cls, client: Pararamio) -> list[Team]:
        """

        Loads teams from the Pararamio client.

        @param client: An instance of the Pararamio client.
        @return: A list of Team objects.
        """
        ids = cls.get_my_team_ids(client)

        if ids:
            url = '/core/org?ids=' + ','.join(map(str, ids))
            res = client.api_get(url, response_model=TeamsResponse)

            if res and 'orgs' in res:
                return [cls(client, **r) for r in res['orgs']]

        return []

    def mark_all_messages_as_read(self) -> bool:
        return self.client.mark_all_messages_as_read(self.id)



"""Read status response schema."""

from __future__ import annotations

from typing import TypedDict


class ReadStatusResponse(TypedDict, total=False):
    """Response from /core/chat/{id}/read_status endpoint."""

    last_read_post_no: int
    posts_count: int
    unread_count: int


__all__ = ['ReadStatusResponse']



"""Core Group model without lazy loading."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Unpack, cast

from pararamio._core.api_schemas.responses import GroupResponseItem
from pararamio._core.exceptions import PararamioValidationError
from pararamio._core.utils.helpers import parse_iso_datetime

from .base import CoreBaseModel

if TYPE_CHECKING:
    from datetime import datetime

    from pararamio._core._types import FormatterT
__all__ = ('CoreGroup',)


# Attribute formatters for Group
GROUP_ATTR_FORMATTERS: FormatterT = {
    'time_created': parse_iso_datetime,
    'time_updated': parse_iso_datetime,
}


class CoreGroup(CoreBaseModel[GroupResponseItem]):
    """Core Group model with common functionality."""

    _data: GroupResponseItem
    # Group attributes
    id: int
    name: str
    slug: str | None
    description: str | None
    info: str | None
    public: bool
    verified: bool
    users_count: int
    time_created: datetime | None
    time_updated: datetime | None

    _attr_formatters: ClassVar[FormatterT] = GROUP_ATTR_FORMATTERS

    def __init__(  # type: ignore[misc]
        self,
        client: Any | None = None,
        id: int | None = None,
        **kwargs: Unpack[GroupResponseItem],
    ) -> None:
        """Initialize group model with data.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            id: Group ID (optional positional or keyword argument)
            **kwargs: Group data
        """
        # Handle positional id
        if id is not None:
            if 'id' in kwargs:
                kwargs.pop('id')
            kwargs['id'] = id
        # Handle ID aliases
        elif 'group_id' in kwargs and 'id' not in kwargs:
            kwargs['id'] = kwargs.pop('group_id')
        self._data = cast('GroupResponseItem', kwargs)

        # Validate that required ID is present
        if 'id' not in self._data:
            raise PararamioValidationError(
                'Group requires id to be present in data. '
                'Cannot create a Group without a valid group identifier.'
            )

        super().__init__(client, **kwargs)  # type: ignore[call-arg]

    def __str__(self) -> str:
        return self.name

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoreGroup):
            return id(other) == id(self)
        return self.id == other.id

    @property
    def members(self) -> list[int]:
        """Get group member user IDs.

        Returns:
            List of user IDs
        """
        users: list[int] = self._data.get('users', [])
        return users if isinstance(users, list) else []



"""Cache protocol definitions for synchronous cache backends."""

# Re-export from core to maintain backward compatibility
from pararamio._core.protocols import CacheProtocol, CacheStatsProtocol

__all__ = ('CacheProtocol', 'CacheStatsProtocol')



from __future__ import annotations

import hashlib
import math
import uuid
from collections.abc import Mapping
from datetime import UTC, datetime
from html import unescape
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
)

from pararamio._core.constants import DATETIME_FORMAT
from pararamio._core.exceptions import PararamioValidationError, PararamModelNotLoadedError

if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    from datetime import timedelta

    from pararamio._core._types import FormatterT

__all__ = (
    'check_login_opts',
    'encode_chat_id',
    'encode_digit',
    'format_datetime',
    'format_or_none',
    'get_empty_vars',
    'get_formatted_attr_or_load',
    'get_utc',
    'join_ids',
    'parse_datetime',
    'parse_iso_datetime',
    'unescape_dict',
)


def check_login_opts(login: str | None, password: str | None) -> bool:
    """
    Check if both login and password options are provided and not empty.

    Parameters:
    login (Optional[str]): The login string to check.
    password (Optional[str]): The password string to check.

    Returns:
    bool: True if both login and password are provided and not empty, False otherwise.
    """
    return all(map(bool, [login, password]))


def get_empty_vars(**kwargs: str) -> str:
    """
    Identifies and returns a comma-separated string of keys from the keyword
    arguments where the corresponding values are empty.

    Parameters:
        **kwargs (Any): Arbitrary keyword arguments with values to be checked.

    Returns:
        str: A comma-separated string of keys with empty values.
    """
    return ', '.join([k for k, v in kwargs.items() if not v])


def encode_digit(digit: int, res: str = '') -> str:
    """
    Encodes a given integer into a custom base-64-like string.

    Parameters:
    digit (int): The integer to be encoded.
    res (str): The resulting encoded string, used internally for recursion.

    Returns:
    str: The encoded string representation of the given integer.
    """
    if not isinstance(digit, int):
        digit = int(digit)
    # noinspection SpellCheckingInspection
    code_string = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_.'
    result = math.floor(digit / len(code_string))
    res = code_string[int(digit % len(code_string))] + res
    return encode_digit(result, res) if result > 0 else res


def encode_chat_id(chat_id: int, posts_count: int, last_read_post_no: int) -> str:
    """
    Encodes the given chat details into a single string.

    This function takes a chat ID, the count of posts in the chat, and the number
    of the last read post, and combines them into a single string separated by hyphens.

    Parameters:
    chat_id (int): The unique identifier for the chat.
    posts_count (int): The total number of posts in the chat.
    last_read_post_no (int): The number of the last post read by the user.

    Returns:
    str: A single string containing the encoded chat details separated by hyphens.
    """
    return '-'.join(map(str, [chat_id, posts_count, last_read_post_no]))


def encode_chats_ids(chats_ids: list[tuple[int, int, int]]) -> str:
    """
    Encodes a list of chat IDs into a single string representation.

    This function takes a list of chat ID tuples and converts each tuple into an encoded string
    using the encode_chat_id function.
    The encoded strings are then joined together by a '/' delimiter.

    Parameters:
        chats_ids (List[Tuple[int, int, int]]): A list of tuples, where each tuple contains
        three integers representing a chat ID.

    Returns:
        str: A single string containing all encoded chat IDs joined by '/'.
    """
    return '/'.join(encode_chat_id(*chat_id) for chat_id in chats_ids)


def join_ids(items: Sequence[Any]) -> str:
    """
    Converts a list of items into a single comma-separated string.

    Args:
        items (List[Any]): A list containing elements of any type to be joined.

    Returns:
        str: A comma-separated string representation of the elements in the list.
    """
    return ','.join(map(str, items))


def get_utc(date: datetime) -> datetime:
    """
    Converts an offset-aware datetime object to its UTC equivalent.

    Parameters:
    date (datetime): The datetime object to convert.
    """
    if date.tzinfo is None:
        msg = 'is not offset-aware datetime'
        raise PararamioValidationError(msg)
    return date - cast('timedelta', date.utcoffset())


def parse_datetime(
    data: Mapping[str, Any],
    key: str,
    format_: str = DATETIME_FORMAT,
) -> datetime | None:
    """
    Parse a datetime object from a dictionary.

    Parameters:
    data (Dict[str, Any]): The dictionary containing the datetime string to parse.
    key (str): The key in the dictionary where the datetime string is stored.
    format_ (str): The format in which the datetime string is stored.

    Returns:
    datetime | None: The parsed datetime object with UTC timezone, or None if value is None.

    Raises:
        KeyError: If the key is not found in data.
    """
    value = data[key]  # Raises KeyError if key not in data
    if value is None:
        return None
    return datetime.strptime(value, format_).replace(tzinfo=UTC)


def parse_iso_datetime(data: Mapping[str, Any], key: str) -> datetime | None:
    """
    Parses an ISO 8601 formatted datetime string from a dictionary by a given key.

    Parameters:
    data (Dict[str, Any]): The dictionary containing the datetime string.
    key (str): The key used to extract the datetime string from the dictionary.

    Returns:
    Optional[datetime]: A datetime object if parsing is successful, otherwise None.
    """
    try:
        return parse_datetime(data, key, '%Y-%m-%dT%H:%M:%S.%fZ')
    except ValueError:
        return parse_datetime(data, key, '%Y-%m-%dT%H:%M:%SZ')


def format_datetime(date: datetime) -> str:
    """
    Formats the given datetime object to a string in UTC using a predefined format.

    Arguments:
        date (datetime): The datetime object to format.

    Returns:
        str: The formatted datetime string.
    """
    return get_utc(date).strftime(DATETIME_FORMAT)


def rand_id() -> str:
    """

    Generates a pseudo-random identifier.

    This function generates a pseudo-random identifier by using the UUID
    and MD5 hashing algorithm. The UUID is first converted to its hexadecimal
    representation and encoded into bytes.
    An MD5 hash is then computed from these bytes.
    The
    resulting hash is converted to an integer,
    scaled by a factor of 10^-21, and finally returned as a string.

    Returns:
        str: The generated pseudo-random identifier
    """
    _hash = hashlib.md5(bytes(uuid.uuid4().hex, 'utf8'))
    return str(int(int(_hash.hexdigest(), 16) * 10**-21))


T = TypeVar('T')


def unescape_dict(d: T, keys: list[str]) -> T:
    """
    Unescapes the values of specified keys in a dictionary.

    This function takes a dictionary and a list of keys,
    and returns a new dictionary where the values of the specified keys have been unescaped.
    All other keys and values remain unchanged.

    Args:
        d: The dictionary whose values are to be unescaped.
        keys: A list of keys for which the values should be unescaped.

    Returns:
        A new dictionary with the values of specified keys unescaped.
    """
    return cast(
        'T',
        {
            k: unescape(v) if k in keys and v is not None else v
            for k, v in cast('dict[str, Any]', d).items()
        },
    )


def format_or_none(key: str, data: dict[str, Any], formatter: FormatterT | None) -> Any:
    """
    Formats the value associated with the given key if a formatter is provided;
    otherwise, returns the unformatted value.

    Parameters:
    key (str): The key for which the value should be retrieved and optionally formatted.
    data (Dict[str, Any]): The dictionary containing the data.
    formatter (Optional[FormatterT]): An optional formatter dictionary where keys are the same as
                                      in data and values are formatting functions.

    Returns:
    Any: The formatted value associated with the key if a formatter exists for it;
    otherwise, the unformatted value.
    """
    if formatter is not None and key in formatter:
        return formatter[key](data, key)
    return data[key]


def get_formatted_attr_or_load(
    obj: object,
    key: str,
    formatter: FormatterT | None = None,
    load_fn: Callable[[], Any] | None = None,
) -> Any:
    """
    Fetches a formatted attribute from an object's `_data` attribute if it exists,
    using an optional formatter function. If the attribute does not exist and a
    `load_fn` function is provided, the function will be called to load the data,
    and the attribute will be fetched again.

    Args:
        obj (object): The object containing the `_data` attribute.
        key (str): The key to look up in the `_data` attribute.
        formatter (Optional[FormatterT], optional): An optional formatter function to
            format the retrieved attribute. Defaults to None.
        load_fn (Optional[Callable[[], Any]], optional): An optional function to call
            if the key does not exist in the `_data` attribute. Defaults to None.

    Returns:
        Any: The formatted attribute if found and formatted, else raises PararamModelNotLoadedError.

    Raises:
        PararamModelNotLoadedError: If the key does not exist in the `_data` attribute and no
            `load_fn` is provided.
    """
    try:
        return format_or_none(key, getattr(obj, '_data', {}), formatter)
    except KeyError as e:
        if load_fn is not None:
            load_fn()
            return format_or_none(key, getattr(obj, '_data', {}), formatter)
        msg = f"Attribute '{key}' has not been loaded yet"
        raise PararamModelNotLoadedError(msg) from e



"""Constants module."""

from .base import (
    BASE_API_URL,
    DATETIME_FORMAT,
    DEFAULT_CONNECTION_NAME,
    FILE_UPLOAD_URL,
    POSTS_LIMIT,
    REQUEST_TIMEOUT,
    UPLOAD_TIMEOUT,
    VERSION,
    XSRF_HEADER_NAME,
)
from .endpoints import (
    AUTH_ENDPOINTS,
    AUTH_INIT_URL,
    AUTH_LOGIN_URL,
    AUTH_NEXT_URL,
    AUTH_TOTP_URL,
    CHAT_ENDPOINTS,
    FILE_ENDPOINTS,
    POST_ENDPOINTS,
    USER_ENDPOINTS,
)

__all__ = [
    'AUTH_ENDPOINTS',
    'AUTH_INIT_URL',
    'AUTH_LOGIN_URL',
    'AUTH_NEXT_URL',
    'AUTH_TOTP_URL',
    'BASE_API_URL',
    'CHAT_ENDPOINTS',
    'DATETIME_FORMAT',
    'DEFAULT_CONNECTION_NAME',
    'FILE_ENDPOINTS',
    'FILE_UPLOAD_URL',
    'POSTS_LIMIT',
    'POST_ENDPOINTS',
    'REQUEST_TIMEOUT',
    'UPLOAD_TIMEOUT',
    'USER_ENDPOINTS',
    'VERSION',
    'XSRF_HEADER_NAME',
]



"""Chat API response schemas."""

from __future__ import annotations

from typing import Any, TypedDict


class ChatResponseItem(TypedDict, total=False):
    """Schema for a single Chat item in API response."""

    # IDs
    id: int
    chat_id: int

    # Type
    type: str  # 'pm' | 'e2e' | 'info' | 'group'

    # Basic info
    title: str
    description: str | None
    custom_title: str | None

    # Type and mode
    history_mode: str  # 'all' | 'since_join'
    pm: bool  # DEPRECATED - Is private message
    e2e: bool  # DEPRECATED - End-to-end encrypted

    # Counts
    posts_count: int
    history_start: int
    last_read_post_no: int

    # Timestamps (as ISO strings from API)
    time_created: str
    time_updated: str
    time_edited: str
    user_time_edited: str | None

    # User IDs
    author_id: int | None
    inviter_id: int | None
    last_msg_author_id: int

    # Organization
    organization_id: int | None
    org_visible: bool

    # Settings
    two_step_required: bool
    posts_live_time: str | None  # timedelta-sec format
    allow_api: bool  # DEPRECATED
    read_only: bool
    parent_id: int | None
    is_common: bool
    is_voice: bool

    # State flags
    tnew: bool
    adm_flag: bool  # DEPRECATED
    is_favorite: bool
    tshow: bool
    mute: str | None  # 'group' | 'total' | None

    # Lists
    pinned: list[int]
    thread_groups: list[int]
    thread_users: list[int]
    thread_admins: list[int]
    thread_users_all: list[int]
    thread_guests: list[int]

    # Last message info
    last_msg_author: str | None  # DEPRECATED
    last_msg_bot_name: str | None  # DEPRECATED
    last_msg_text: str | None  # DEPRECATED
    last_msg: str

    # Additional fields
    meta: dict[str, Any] | None  # DEPRECATED
    keys: dict[str, str] | None  # Only for e2e chats


class ChatsResponse(TypedDict):
    """Schema for Chats API response wrapper."""

    chats: list[ChatResponseItem]


class ChatSearchResponse(TypedDict):
    """Schema for /core/chat/search API response."""

    flt: str
    threads: list[ChatResponseItem]


class KeywordsResponse(TypedDict, total=False):
    """Response from chat keywords endpoint."""

    kw: str  # Keywords' string (optional)


class ChatStatusItem(TypedDict):
    """Status info for a single chat in sync response."""

    posts_count: int  # Total number of posts
    last_read_post_no: int  # Last read post-number
    last_msg: str  # Last message text
    time_updated: str  # ISO datetime string


class ChatSyncResponse(TypedDict):
    """Response from POST /core/chat/sync endpoint."""

    new: list[int]  # New chat IDs
    updated: list[int]  # Updated chat IDs
    removed: list[int]  # Removed chat IDs
    status: dict[str, ChatStatusItem]  # Status updates keyed by chat ID


class ChatListResponse(TypedDict):
    """Response from GET /core/chat/sync endpoint."""

    chats: list[int]  # List of chat IDs


__all__ = [
    'ChatListResponse',
    'ChatResponseItem',
    'ChatSearchResponse',
    'ChatStatusItem',
    'ChatSyncResponse',
    'ChatsResponse',
    'KeywordsResponse',
]



"""Universal lazy loading utilities for posts iteration."""

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class LazyLoadingConfig:
    """Configuration for lazy loading with cache-first approach.

    Attributes:
        start_post_no: Starting post number (1-based)
        end_post_no: Ending post number (inclusive)
        per_request: Number of posts to load per batch
        cache_key_template: Template for cache keys, e.g. "chat.{chat_id}.post.{post_no}"
        chat_id: Chat ID for cache key generation
    """

    start_post_no: int
    end_post_no: int
    per_request: int
    cache_key_template: str
    chat_id: int


@dataclass
class LazyLoadBatch:
    """Represents a batch of posts to load.

    Attributes:
        start: Starting post number
        end: Ending post number (inclusive)
        retry_count: Number of retry attempts for this batch
    """

    start: int
    end: int
    retry_count: int = 0


def generate_cache_key(template: str, chat_id: int, post_no: int) -> str:
    """Generate a cache key for a specific post.

    Args:
        template: Cache key template with placeholders
        chat_id: Chat ID
        post_no: Post number

    Returns:
        Formatted cache key
    """
    return template.format(chat_id=chat_id, post_no=post_no)


def get_retry_delay(attempt: int) -> float:
    """Calculate retry delay with exponential backoff.

    Args:
        attempt: Attempt number (0-based)

    Returns:
        Delay in seconds: 0.5s, 1s, 2s for attempts 0, 1, 2
    """
    return float(0.5 * (2**attempt))



"""Post-event schemas."""

from __future__ import annotations

from typing import Any, Literal, TypedDict


class UserInfo(TypedDict):
    """User info in events."""

    id: int
    name: str
    unique_name: str


class GroupInfo(TypedDict):
    """Group info in events."""

    id: int
    name: str
    unique_name: str | None


class TeamInfo(TypedDict):
    """Team/Organization info in events."""

    id: int
    title: str


class ChatInfo(TypedDict):
    """Chat info in events."""

    id: int
    title: str
    posts_qty: int | None


# Event data types for different event types


class NewThreadEventData(TypedDict):
    """NEW_THREAD event data."""

    users_add: list[UserInfo]
    groups_add: list[GroupInfo]


class EditThreadEventData(TypedDict):
    """EDIT_THREAD event data."""

    users_add: list[UserInfo]
    users_del: list[UserInfo]
    admins_add: list[UserInfo]
    admins_del: list[UserInfo]
    groups_add: list[GroupInfo]
    groups_del: list[GroupInfo]


class EnterToThreadEventData(TypedDict):
    """ENTER_TO_THREAD event data."""

    user_id: int


class QuitFromThreadEventData(TypedDict):
    """QUIT_FROM_THREAD event data."""

    user: UserInfo
    groups_exist: list[dict[str, Any]]


class GroupsQuitFromThreadEventData(TypedDict):
    """GROUPSQUIT_FROM_THREAD event data."""

    __root__: list[GroupInfo]


class WelcomeToOrgEventData(TypedDict):
    """WELCOME_TO_ORG event data."""

    organization: TeamInfo
    welcome_text: str


class ForkThreadEventData(TypedDict):
    """FORK_THREAD event data."""

    to_chat: ChatInfo
    from_chat: dict[str, Any]  # Contains id, title, posts


class ChatTitleEventData(TypedDict):
    """CHAT_TITLE event data."""

    new_title: str
    old_title: str


class TransferChatEventData(TypedDict):
    """TRANSFER_CHAT event data."""

    chat: ChatInfo
    to_team: TeamInfo | None
    from_team: TeamInfo | None


class UserDeactivatedEventData(TypedDict):
    """USER_DEACTIVATED event data."""

    user: UserInfo


class NewCommonChatEventData(TypedDict):
    """NEW_COMMON_CHAT event data."""

    # Empty data


class DelCommonChatEventData(TypedDict):
    """DEL_COMMON_CHAT event data."""

    # Empty data


class ChangeSettingsEventData(TypedDict):
    """CHANGE_SETTINGS event data."""

    user_id: int
    change_settings: dict[str, Any]


class NewBranchChatEventData(TypedDict):
    """NEW_BRANCH_CHAT event data."""

    chat_id: int
    parent_id: int
    post_no: int


class ConversionBranchChatEventData(TypedDict):
    """CONVERSION_BRANCH_CHAT event data."""

    chat_id: int
    parent_id: int
    post_no: int


class PostPinnedEventData(TypedDict):
    """POST_PINNED/POST_UNPINNED event data."""

    user_unique_name: str
    user_id: int
    post_no: int


class GroupEventData(TypedDict):
    """GROUP_CREATED/GROUP_UPDATED/GROUP_DELETED/GROUP_LEAVED event data."""

    user: UserInfo
    group: GroupInfo
    role: Literal['user', 'admin', 'delete'] | None
    users: list[UserInfo] | None


class OrgQuitEventData(TypedDict):
    """ORG_QUIT event data."""

    user: UserInfo
    organization: TeamInfo


class OrgMembersEventData(TypedDict):
    """ORG_MEMBERS event data."""

    add_users: list[UserInfo]
    del_users: list[UserInfo]
    organization: TeamInfo


# Main event structure


class PostEvent(TypedDict):
    """Post-event structure."""

    type: Literal[
        'NEW_THREAD',
        'EDIT_THREAD',
        'ENTER_TO_THREAD',
        'QUIT_FROM_THREAD',
        'GROUPSQUIT_FROM_THREAD',
        'WELCOME_TO_ORG',
        'FORK_THREAD',
        'CHAT_TITLE',
        'TRANSFER_CHAT',
        'USER_DEACTIVATED',
        'NEW_COMMON_CHAT',
        'DEL_COMMON_CHAT',
        'CHANGE_SETTINGS',
        'NEW_BRANCH_CHAT',
        'CONVERSION_BRANCH_CHAT',
        'POST_PINNED',
        'POST_UNPINNED',
        'GROUP_CREATED',
        'GROUP_UPDATED',
        'GROUP_DELETED',
        'GROUP_LEAVED',
        'ORG_QUIT',
        'ORG_MEMBERS',
    ]
    data: (
        NewThreadEventData
        | EditThreadEventData
        | EnterToThreadEventData
        | QuitFromThreadEventData
        | list[GroupInfo]
        | WelcomeToOrgEventData
        | ForkThreadEventData
        | ChatTitleEventData
        | TransferChatEventData
        | UserDeactivatedEventData
        | dict[str, Any]  # For empty events
        | ChangeSettingsEventData
        | NewBranchChatEventData
        | ConversionBranchChatEventData
        | PostPinnedEventData
        | GroupEventData
        | OrgQuitEventData
        | OrgMembersEventData
    )


__all__ = [
    'ChangeSettingsEventData',
    'ChatInfo',
    'ChatTitleEventData',
    'ConversionBranchChatEventData',
    'DelCommonChatEventData',
    'EditThreadEventData',
    'EnterToThreadEventData',
    'ForkThreadEventData',
    'GroupEventData',
    'GroupInfo',
    'GroupsQuitFromThreadEventData',
    'NewBranchChatEventData',
    'NewCommonChatEventData',
    'NewThreadEventData',
    'OrgMembersEventData',
    'OrgQuitEventData',
    'PostEvent',
    'PostPinnedEventData',
    'QuitFromThreadEventData',
    'TeamInfo',
    'TransferChatEventData',
    'UserDeactivatedEventData',
    'UserInfo',
    'WelcomeToOrgEventData',
]



from __future__ import annotations

import binascii
import json
import logging
import time
from typing import TYPE_CHECKING, Any
from urllib.error import HTTPError

from pararamio._core import exceptions as ex
from pararamio._core.constants import XSRF_HEADER_NAME
from pararamio._core.exceptions import RateLimitError
from pararamio._core.utils.auth_flow import generate_otp
from pararamio._core.utils.http_client import RateLimitHandler
from pararamio._core.utils.logging_config import (
    LoggerManager,
    get_logger,
)

try:
    from pararamio._core.utils.captcha import show_captcha

    HAS_CAPTCHA = True
except ImportError:
    HAS_CAPTCHA = False
    show_captcha = None

from .requests import api_request, raw_api_request

if TYPE_CHECKING:
    from http.cookiejar import CookieJar

    from pararamio._core._types import HeaderLikeT, SecondStepFnT

__all__ = (
    'authenticate',
    'do_second_step',
    'do_second_step_with_code',
    'get_xsrf_token',
)

XSRF_URL = INIT_URL = '/auth/init'
LOGIN_URL = '/auth/login/password'
TWO_STEP_URL = '/auth/totp'
AUTH_URL = '/auth/next'

# Get component-specific loggers
log = logging.getLogger('pararamio')  # Keep for backward compatibility
auth_logger = get_logger(LoggerManager.AUTH)
rate_limit_logger = get_logger(LoggerManager.RATE_LIMIT)
session_logger = get_logger(LoggerManager.SESSION)


def get_xsrf_token(cookie_jar: CookieJar) -> str:
    auth_logger.debug('Requesting XSRF token from %s', XSRF_URL)
    _, headers = raw_api_request(XSRF_URL, cookie_jar=cookie_jar)
    for key, value in headers:
        if key.lower() == 'x-xsrftoken':
            # Only log first 8 chars for security
            auth_logger.debug('XSRF token obtained: %s...', value[:8] if len(value) > 8 else '***')
            return value
    auth_logger.error('XSRF token not found in response headers')
    msg = f'XSRF Header was not found in {XSRF_URL} url'
    raise ex.PararamioXSRFRequestError(msg)


def do_init(cookie_jar: CookieJar, headers: dict[str, str]) -> tuple[bool, dict[str, Any]]:
    auth_logger.debug('Initializing authentication session')
    try:
        result = api_request(
            INIT_URL,
            method='GET',
            headers=headers,
            cookie_jar=cookie_jar,
        )
        auth_logger.debug('Authentication session initialized successfully')
        return True, result
    except HTTPError as e:
        if e.code < 500:
            auth_logger.warning('Session init failed with status %d', e.code)
            return False, json.loads(e.read())
        auth_logger.error('Session init server error: %d', e.code)
        raise


def do_login(
    login: str, password: str, cookie_jar: CookieJar, headers: dict[str, str]
) -> tuple[bool, dict[str, Any]]:
    auth_logger.info('Attempting login for user: %s', login)
    try:
        result = api_request(
            LOGIN_URL,
            method='POST',
            data={'email': login, 'password': password},
            headers=headers,
            cookie_jar=cookie_jar,
        )
        auth_logger.info('Login successful for user: %s', login)
        return True, result
    except HTTPError as e:
        if e.code < 500:
            auth_logger.warning('Login failed for user %s with status %d', login, e.code)
            return False, json.loads(e.read())
        auth_logger.error('Login server error for user %s: %d', login, e.code)
        raise


def do_taking_secret(cookie_jar: CookieJar, headers: dict[str, str]) -> tuple[bool, dict[str, Any]]:
    try:
        return True, api_request(
            AUTH_URL,
            method='GET',
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise


def do_second_step(
    cookie_jar: CookieJar, headers: dict[str, str], key: str
) -> tuple[bool, dict[str, str]]:
    """
    do second step pararam login with a TFA key or raise Exception
    :param cookie_jar: cookie container
    :param headers: headers to send
    :param key: key to generate one time code
    :return: True if login success
    """
    auth_logger.debug('Starting 2FA with TOTP key')
    if not key:
        auth_logger.error('2FA key is empty')
        msg = 'key can not be empty'
        raise ex.PararamioSecondFactorAuthenticationError(msg)
    try:
        key = generate_otp(key)
        auth_logger.debug('Generated OTP code from key')
    except binascii.Error as e:
        auth_logger.error('Invalid 2FA key format')
        msg = 'Invalid second step key'
        raise ex.PararamioSecondFactorAuthenticationError(msg) from e
    try:
        resp = api_request(
            TWO_STEP_URL,
            method='POST',
            data={'code': key},
            headers=headers,
            cookie_jar=cookie_jar,
        )
        auth_logger.debug('2FA code accepted')
    except HTTPError as e:
        if e.code < 500:
            auth_logger.warning('2FA failed with status %d', e.code)
            return False, json.loads(e.read())
        auth_logger.error('2FA server error: %d', e.code)
        raise
    return True, resp


def do_second_step_with_code(
    cookie_jar: CookieJar, headers: dict[str, str], code: str
) -> tuple[bool, dict[str, str]]:
    """
    do second step pararam login with TFA code or raise Exception
    :param cookie_jar: cookie container
    :param headers: headers to send
    :param code: 6 digits code
    :return:  True if login success
    """
    if not code:
        msg = 'code can not be empty'
        raise ex.PararamioSecondFactorAuthenticationError(msg)
    if len(code) != 6:
        msg = 'code must be 6 digits len'
        raise ex.PararamioSecondFactorAuthenticationError(msg)
    try:
        resp = api_request(
            TWO_STEP_URL,
            method='POST',
            data={'code': code},
            headers=headers,
            cookie_jar=cookie_jar,
        )
    except HTTPError as e:
        if e.code < 500:
            return False, json.loads(e.read())
        raise
    return True, resp


def _handle_rate_limit(wait_auth_limit: bool) -> None:
    """Handle rate limiting before authentication."""
    rate_limit_handler = RateLimitHandler()
    should_wait, wait_seconds = rate_limit_handler.should_wait()
    if should_wait:
        if wait_auth_limit:
            rate_limit_logger.warning(
                'Rate limit active, waiting %d seconds before authentication', wait_seconds
            )
            time.sleep(wait_seconds)
            rate_limit_logger.debug('Rate limit wait completed')
        else:
            rate_limit_logger.error(
                'Rate limit exceeded, would need to wait %d seconds', wait_seconds
            )
            msg = f'Rate limit exceeded. Retry after {wait_seconds} seconds'
            raise RateLimitError(msg, retry_after=wait_seconds)


def _handle_captcha(
    login: str, password: str, cookie_jar: CookieJar, headers: dict[str, Any], resp: dict[str, Any]
) -> tuple[bool, tuple[bool, dict[str, Any]]]:
    """Handle captcha requirement during authentication.

    Returns:
        Tuple of (was_captcha_required, (success, response))
    """
    if resp.get('codes', {}).get('non_field', '') != 'captcha_required':
        return False, (False, resp)

    auth_logger.info('Captcha required for user: %s', login)

    if not HAS_CAPTCHA:
        auth_logger.error('Captcha module not available')
        msg = 'Captcha required, but captcha module not available'
        raise ex.PararamioCaptchaAuthenticationError(msg)

    # show_captcha is guaranteed to be callable here (HAS_CAPTCHA is True)
    assert show_captcha is not None
    auth_logger.debug('Showing captcha for user: %s', login)
    success = show_captcha(f'login:{login}', headers, cookie_jar)
    if not success:
        auth_logger.error('Captcha verification failed for user: %s', login)
        msg = 'Captcha required'
        raise ex.PararamioCaptchaAuthenticationError(msg)

    auth_logger.info('Captcha solved, retrying login for user: %s', login)
    return True, do_login(login, password, cookie_jar, headers)


def authenticate(
    login: str,
    password: str,
    cookie_jar: CookieJar,
    headers: HeaderLikeT | None = None,
    second_step_fn: SecondStepFnT | None = do_second_step,
    second_step_arg: str | None = None,
    wait_auth_limit: bool = False,
) -> tuple[bool, dict[str, Any], str]:
    auth_logger.info(
        'Starting authentication for user: %s (2FA: %s)',
        login,
        'enabled' if second_step_arg else 'disabled',
    )
    start_time = time.perf_counter()

    # Handle rate limiting
    _handle_rate_limit(wait_auth_limit)

    if not headers or XSRF_HEADER_NAME not in headers:
        if headers is None:
            headers = {}
        auth_logger.debug('XSRF token not in headers, fetching new one')
        headers[XSRF_HEADER_NAME] = get_xsrf_token(cookie_jar)

    success, resp = do_login(login, password, cookie_jar, headers)

    # Handle captcha if required
    captcha_handled, captcha_result = _handle_captcha(login, password, cookie_jar, headers, resp)
    if captcha_handled:
        success, resp = captcha_result

    if not success and resp.get('error', 'xsrf'):
        log.debug('invalid xsrf trying to get new one')
        auth_logger.warning('XSRF token invalid, fetching new one')
        headers[XSRF_HEADER_NAME] = get_xsrf_token(cookie_jar)
        auth_logger.info('Retrying login with new XSRF token')
        success, resp = do_login(login, password, cookie_jar, headers)

    if not success:
        error = resp.get('error', 'unknown')
        log.error('authentication failed: %s', error)
        auth_logger.error('Authentication failed for user %s: %s', login, error)
        msg = 'Login, password authentication failed'
        raise ex.PararamioPasswordAuthenticationError(msg)

    if second_step_fn is not None and second_step_arg:
        auth_logger.info('Performing second factor authentication')
        success, resp = second_step_fn(cookie_jar, headers, second_step_arg)
        if not success:
            auth_logger.error('Second factor authentication failed for user: %s', login)
            msg = 'Second factor authentication failed'
            raise ex.PararamioSecondFactorAuthenticationError(msg)
        auth_logger.info('Second factor authentication successful')

    auth_logger.debug('Completing authentication flow')
    success, resp = do_taking_secret(cookie_jar, headers)
    if not success:
        auth_logger.error('Failed to obtain authentication secret')
        msg = 'Taking secret failed'
        raise ex.PararamioAuthenticationError(msg)

    success, resp = do_init(cookie_jar, headers)

    elapsed_ms = (time.perf_counter() - start_time) * 1000
    auth_logger.info(
        'Authentication completed successfully for user %s in %.2fms', login, elapsed_ms
    )

    # Log session establishment
    session_logger.info(
        'Session established for user: %s (user_id: %s)', login, resp.get('user_id', 'unknown')
    )

    return True, {'user_id': resp.get('user_id')}, headers[XSRF_HEADER_NAME]



"""Common API response schemas used across multiple endpoints."""

from __future__ import annotations

from typing import TypedDict


class OkResponse(TypedDict):
    """Response with result: OK status."""

    result: str  # Should be "OK"


class EmptyResponse(TypedDict, total=False):
    """Empty response (empty dict)."""


class ChatIdResponse(TypedDict):
    """Response containing chat_id."""

    chat_id: int


class GenericResponse(TypedDict, total=False):
    """Generic response that can contain any fields."""

    # This allows any additional fields
    # Use when the exact response structure is dynamic


__all__ = [
    'ChatIdResponse',
    'EmptyResponse',
    'GenericResponse',
    'OkResponse',
]



"""Performance monitoring utilities and decorators."""

from __future__ import annotations

import contextlib
import functools
import logging
import time
from collections.abc import Awaitable, Callable
from typing import Any, TypeVar, cast

from pararamio._core.utils.logging_config import LoggerManager, get_logger

# Get component-specific loggers
cache_logger = get_logger(LoggerManager.CACHE)
http_logger = get_logger(LoggerManager.HTTP_CLIENT)
lazy_logger = get_logger(LoggerManager.LAZY_LOADING)
batch_logger = get_logger(LoggerManager.BATCH_LOGIC)

T = TypeVar('T')


def _log_operation_start(
    logger: Any, operation: str, log_args: bool, args: tuple[Any, ...]
) -> None:
    """Log the start of an operation."""
    if log_args and args:
        # Try to extract key from first arg if it looks like a self.method(key, ...) call
        if len(args) > 1 and hasattr(args[0], '__class__'):
            # Skip 'self' argument
            key_arg = args[1] if len(args) > 1 else None
            if key_arg:
                logger.debug('%s START: key=%s', operation, key_arg)
        else:
            logger.debug('%s START: args=%s', operation, args[:2])  # Limit args logging
    else:
        logger.debug('%s START', operation)


def _log_operation_result(
    logger: Any,
    operation: str,
    result: Any,
    elapsed_ms: float,
    log_result: bool,
) -> None:
    """Log the result of an operation."""
    # Determine if it was a hit or miss for cache operations
    if 'cache' in operation.lower():
        if result is None:
            logger.debug('%s MISS: elapsed=%.2fms', operation, elapsed_ms)
        else:
            # Try to get size if result has __len__
            try:
                size = len(result) if hasattr(result, '__len__') else 0
                if size > 0:
                    logger.debug('%s HIT: size=%d, elapsed=%.2fms', operation, size, elapsed_ms)
                else:
                    logger.debug('%s HIT: elapsed=%.2fms', operation, elapsed_ms)
            except Exception:
                logger.debug('%s HIT: elapsed=%.2fms', operation, elapsed_ms)
    else:
        # For non-cache operations
        if log_result and result is not None:
            logger.debug(
                '%s COMPLETE: result=%s, elapsed=%.2fms',
                operation,
                str(result)[:100],  # Limit result logging
                elapsed_ms,
            )
        else:
            logger.debug('%s COMPLETE: elapsed=%.2fms', operation, elapsed_ms)


def monitor_performance(
    operation: str,
    logger: Any | None = None,
    log_args: bool = True,
    log_result: bool = False,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Monitor performance of a synchronous function.

    Args:
        operation: Name of the operation being monitored.
        logger: Logger to use (defaults to cache_logger).
        log_args: Whether to log function arguments.
        log_result: Whether to log function result.

    Returns:
        Decorated function with performance monitoring.
    """
    if logger is None:
        logger = cache_logger

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            # Check if logging is enabled to avoid measurement overhead
            if not logger.isEnabledFor(logging.DEBUG):
                return func(*args, **kwargs)

            start_time = time.perf_counter()
            _log_operation_start(logger, operation, log_args, args)

            try:
                result = func(*args, **kwargs)
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                _log_operation_result(logger, operation, result, elapsed_ms, log_result)
                return result

            except Exception as e:
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                logger.error('%s ERROR: %s, elapsed=%.2fms', operation, str(e), elapsed_ms)
                raise

        return cast('Callable[..., T]', wrapper)

    return decorator


def monitor_async_performance(
    operation: str,
    logger: Any | None = None,
    log_args: bool = True,
    log_result: bool = False,
) -> Callable[[Callable[..., Awaitable[T]]], Callable[..., Awaitable[T]]]:
    """Monitor performance of an asynchronous function.

    Args:
        operation: Name of the operation being monitored.
        logger: Logger to use (defaults to cache_logger).
        log_args: Whether to log function arguments.
        log_result: Whether to log function result.

    Returns:
        Decorated async function with performance monitoring.
    """
    if logger is None:
        logger = cache_logger

    def decorator(func: Callable[..., Awaitable[T]]) -> Callable[..., Awaitable[T]]:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> T:
            # Check if logging is enabled to avoid measurement overhead
            if not logger.isEnabledFor(logging.DEBUG):
                return await func(*args, **kwargs)

            start_time = time.perf_counter()
            _log_operation_start(logger, operation, log_args, args)

            try:
                result = await func(*args, **kwargs)
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                _log_operation_result(logger, operation, result, elapsed_ms, log_result)
                return result

            except Exception as e:
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                logger.error('%s ERROR: %s, elapsed=%.2fms', operation, str(e), elapsed_ms)
                raise

        return cast('Callable[..., Awaitable[T]]', wrapper)

    return decorator


def monitor_batch_operation(
    operation: str,
    batch_size_extractor: Callable[..., int] | None = None,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Monitor performance of batch operations.

    Args:
        operation: Name of the batch operation.
        batch_size_extractor: Function to extract batch size from arguments.

    Returns:
        Decorated function with batch operation monitoring.
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            # Check if logging is enabled to avoid measurement overhead
            if not batch_logger.isEnabledFor(logging.DEBUG):
                return func(*args, **kwargs)

            start_time = time.perf_counter()

            # Extract batch size if possible
            batch_size = 0
            if batch_size_extractor:
                with contextlib.suppress(Exception):
                    batch_size = batch_size_extractor(*args, **kwargs)

            if batch_size > 0:
                batch_logger.debug('%s START: batch_size=%d', operation, batch_size)
            else:
                batch_logger.debug('%s START', operation)

            try:
                result = func(*args, **kwargs)
                elapsed_ms = (time.perf_counter() - start_time) * 1000

                if batch_size > 0:
                    batch_logger.info(
                        '%s COMPLETE: batch_size=%d, elapsed=%.2fms (%.2fms/item)',
                        operation,
                        batch_size,
                        elapsed_ms,
                        elapsed_ms / batch_size if batch_size > 0 else 0,
                    )
                else:
                    batch_logger.info('%s COMPLETE: elapsed=%.2fms', operation, elapsed_ms)

                return result

            except Exception as e:
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                batch_logger.error('%s ERROR: %s, elapsed=%.2fms', operation, str(e), elapsed_ms)
                raise

        return cast('Callable[..., T]', wrapper)

    return decorator


def monitor_http_request(
    method: str | None = None,
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Monitor performance of HTTP requests.

    Args:
        method: HTTP method (GET, POST, etc.). If None, will try to extract from args.

    Returns:
        Decorated function with HTTP request monitoring.
    """

    def decorator(func: Callable[..., T]) -> Callable[..., T]:
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> T:
            # Check if logging is enabled to avoid measurement overhead
            if not http_logger.isEnabledFor(logging.DEBUG):
                return func(*args, **kwargs)

            start_time = time.perf_counter()

            # Try to extract URL and method from args
            url = kwargs.get('url', args[1] if len(args) > 1 else 'unknown')
            http_method = method or kwargs.get('method', 'GET')

            http_logger.debug('HTTP %s %s START', http_method, url)

            try:
                result = func(*args, **kwargs)
                elapsed_ms = (time.perf_counter() - start_time) * 1000

                # Try to get status code from result if it's a response object
                status_code = None
                if hasattr(result, 'status_code'):
                    status_code = result.status_code
                elif hasattr(result, 'status'):
                    status_code = result.status

                if status_code:
                    http_logger.info(
                        'HTTP %s %s: status=%d, elapsed=%.2fms',
                        http_method,
                        url,
                        status_code,
                        elapsed_ms,
                    )
                else:
                    http_logger.info('HTTP %s %s: elapsed=%.2fms', http_method, url, elapsed_ms)

                return result

            except Exception as e:
                elapsed_ms = (time.perf_counter() - start_time) * 1000
                http_logger.error(
                    'HTTP %s %s ERROR: %s, elapsed=%.2fms', http_method, url, str(e), elapsed_ms
                )
                raise

        return cast('Callable[..., T]', wrapper)

    return decorator


# Export all decorators
__all__ = [
    'monitor_async_performance',
    'monitor_batch_operation',
    'monitor_http_request',
    'monitor_performance',
]



"""Bot API response schemas."""

from __future__ import annotations

from typing import Any, TypedDict


class BotMessageResponse(TypedDict):
    """Response from /bot/message and /msg/post/private endpoints."""

    chat_id: int
    post_no: int


class PrivateMessageResponse(TypedDict):
    """Response from /msg/post/private endpoint."""

    chat_id: int
    post_no: int


class TaskStatusResponse(TypedDict):
    """Response from task status endpoint."""

    result: str


class ActivitiesResponse(TypedDict):
    """Response from activity endpoint."""

    activities: list[dict[str, Any]]


__all__ = [
    'ActivitiesResponse',
    'BotMessageResponse',
    'PrivateMessageResponse',
    'TaskStatusResponse',
]



from __future__ import annotations

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, Unpack
from uuid import uuid4

# Imports from core
from pararamio._core import (
    PararamioRequestError,
    PararamioTypeError,
    PararamioValidationError,
    PararamMultipleFoundError,
    PararamNotFoundError,
)
from pararamio._core.api_schemas import (
    DataListResponse,
    PostCreateResponse,
    PostDeleteResponse,
    PostEditResponse,
    PostResponseItem,
    PostsResponse,
    RerereResponse,
    WhoReadResponse,
)
from pararamio._core.models import CoreFile as File
from pararamio._core.models import CorePost
from pararamio._core.utils.helpers import (
    encode_digit,
    rand_id,
)
from pararamio._core.utils.ranges import combine_ranges

from .base import SyncClientMixin

if TYPE_CHECKING:
    from pararamio._core._types import PostMention

    from pararamio.client import Pararamio
    from pararamio.models.chat import Chat

__all__ = ('Post',)


def get_post_mention(item: Any) -> PostMention | None:
    """Convert dict to PostMention if it has required fields.

    Args:
        item: Dictionary with mention data

    Returns:
        PostMention or None if missing required fields
    """
    if 'id' not in item or 'name' not in item or 'value' not in item:
        return None
    return {
        'id': item['id'],
        'name': item['name'],
        'value': item['value'],
    }


class Post(
    CorePost['Chat'],
    SyncClientMixin[PostResponseItem],
):
    """Sync Post model with lazy loading support."""

    def __init__(  # type: ignore[misc]
        self,
        chat: Chat,
        post_no: int | None = None,
        **kwargs: Unpack[PostResponseItem],
    ) -> None:
        """Initialize sync post.

        Args:
            chat: Chat object
            post_no: Post number (optional positional or keyword argument)
            **kwargs: Post data
        """
        super().__init__(chat, post_no, **kwargs)

    @property
    def in_thread_no(self) -> int:
        """Alias for post_no for backward compatibility."""
        return self.post_no

    @property
    def file(self) -> File | None:
        """Get file attachment if present (property for backward compatibility)."""
        return self.get_file()

    def get_file(self) -> File | None:
        """Get attached file if any.

        Returns:
            File object or None if no file
        """
        _file = self.meta.get('file', None)
        if not _file:
            return None
        return File(self._chat._client, **_file)

    def get_attachments(self) -> list[File]:
        """Get all file attachments for this post.

        Returns:
            List of attached files
        """
        attachments = []
        main_file = self.get_file()
        if main_file:
            attachments.append(main_file)
        return attachments

    def download_file(self, filename: str | None = None) -> bytes:
        """Download the file attached to this post.

        Args:
            filename: Optional filename to use for download.
                     If not provided, uses the filename from the file metadata.

        Returns:
            File content as bytes

        Raises:
            PararamioTypeError: If the post doesn't have a file attachment
            PararamioValidationError: If filename cannot be determined
        """
        file = self.file
        if file is None:
            raise PararamioTypeError(f'Post {self.post_no} is not a file post')

        if filename is None:
            filename = getattr(file, 'filename', None) or getattr(file, 'name', None)
            if not filename:
                raise PararamioValidationError('Cannot determine filename')

        bio = self.client.download_file(file.guid, filename)
        return bio.read()

    @property
    def client(self) -> Pararamio:
        """Get client instance."""
        return self._chat._client

    @property
    def is_mention(self) -> bool:
        """Check if current user is mentioned."""
        if not self.text_parsed:
            return False
        for item in self.text_parsed:
            if item.get('type', '') == 'mention' and item.get(
                'id', None
            ) == self.client.get_profile().get('id', -1):
                return True
        return False

    @property
    def mentions(self) -> list[PostMention]:
        """Get post mentions with additional processing."""
        text_parsed = self.text_parsed
        if text_parsed is None:
            return []
        mentions_: list[PostMention] = []
        for item in text_parsed:
            if item.get('type', '') == 'mention':
                mention = get_post_mention(item)
                if not mention:
                    continue
                mentions_.append(mention)
        return mentions_

    @property
    def user_links(self) -> list[PostMention]:
        """Get user links with additional processing."""
        if not self.text_parsed:
            return []
        links: list[PostMention] = []
        for item in self.text_parsed:
            if item.get('type', '') == 'user_link':
                mention = get_post_mention(item)
                if not mention:
                    continue
                links.append(mention)
        return links

    def load(self) -> Post:
        """Load full post data from API with caching.

        Returns:
            Self with updated data

        Raises:
            PararamNotFoundError: If post not found
            PararamMultipleFoundError: If multiple posts found
        """
        # Try to load from cache first if available
        if self.client._cache:
            cache_key = f'chat.{self.chat.id}.post.{self.post_no}'
            cached = self.client._cache.get(cache_key)
            if cached:
                self._data = cached
                return self

        # Load from API if not cached
        url = f'/msg/post?ids={encode_digit(self.chat.id)}-{encode_digit(self.post_no)}'
        response = self.client.api_get(url, response_model=PostsResponse)
        res = response.get('posts', [])
        if len(res) == 0:
            raise PararamNotFoundError(
                f'Post not found: post_no {self.post_no} in chat {self._chat.id}'
            )
        if len(res) > 1:
            raise PararamMultipleFoundError(
                f'Multiple posts found ({len(res)}) for post {self.post_no} in chat {self._chat.id}'
            )
        self._data = res[0]

        # Cache the post data if cache is available
        if self.client._cache:
            cache_key = f'chat.{self.chat.id}.post.{self.post_no}'
            self.client._cache.set(cache_key, self._data)

        return self

    @property
    def replies(self) -> list[int]:
        """Get list of reply post numbers (property for backward compatibility).

        Returns:
            List of post numbers that reply to this post
        """
        return self.get_replies()

    def get_replies(self) -> list[int]:
        """Get list of reply post numbers.

        Returns:
            List of post numbers that reply to this post
        """
        url = f'/msg/post/{self._chat.id}/{self.post_no}/replies'
        response = self.client.api_get(url, response_model=DataListResponse[int])
        return response.get('data', [])

    def load_reply_posts(self) -> list[Post]:
        """Load all posts that reply to this post.

        Returns:
            List of reply posts
        """
        reply_numbers = self.get_replies()
        if not reply_numbers:
            return []

        # Load each reply post
        reply_posts = []
        for post_no in reply_numbers:
            try:
                post = Post(self._chat, post_no=post_no)
                post.load()
                reply_posts.append(post)
            except PararamNotFoundError:
                # Skip posts that were deleted or are not accessible
                continue

        return reply_posts

    def reply(self, text: str, quote: str | None = None) -> Post:
        """Reply to this post.

        Args:
            text: Reply to text
            quote: Optional quote text

        Returns:
            Created reply post
        """
        _url = f'/msg/post/{self._chat.id}'
        res = self.client.api_post(
            _url,
            {'uuid': rand_id(), 'text': text, 'quote': quote, 'reply_no': self.post_no},
            response_model=PostCreateResponse,
        )
        return Post(self._chat, post_no=res['post_no']).load()

    def rerere(self) -> list[Post]:
        """Get all replies in a thread recursively.

        Returns:
            List of all posts in the reply chain
        """
        url = f'/msg/post/{self._chat.id}/{self.post_no}/rerere'
        res = self.client.api_get(url, response_model=RerereResponse)

        post_numbers = res.get('data', [])
        if not post_numbers:
            return []

        # Group post numbers into ranges for batch loading
        ranges = combine_ranges(post_numbers, max_gap=50)

        posts_dict: dict[int, Post] = {}

        # Load posts in batches
        for start, end in ranges:
            if start == end:
                # Single post
                post = Post(self._chat, post_no=start)
                post.load()
                posts_dict[start] = post
            else:
                # Batch load posts in range
                batch_posts = self._chat._load_posts(
                    start_post_no=start,
                    end_post_no=end + 1,  # end_post_no is exclusive in API
                )
                for post in batch_posts:
                    if post.post_no in post_numbers:
                        posts_dict[post.post_no] = post

        # Return posts in the original order from rerere response
        return [posts_dict[post_no] for post_no in post_numbers if post_no in posts_dict]

    def get_tree(self, load_limit: int = 1000) -> OrderedDict[int, Post]:
        """Get post tree with replies.

        Args:
            load_limit: Maximum number of posts to load

        Returns:
            Ordered dictionary of posts in the tree
        """
        posts = {self.post_no: self}
        for post in self.rerere():
            posts[post.post_no] = post
        first = posts[min(posts.keys())]
        tree = OrderedDict(sorted(posts.items()))
        load_start = first.post_no + 1
        if self.post_no - first.post_no > load_limit:
            load_start = self.post_no - load_limit
        for post in self.chat._lazy_posts_loader(*sorted([load_start, self.post_no - 1])):
            posts[post.post_no] = post

        for post in sorted(posts.values()):
            if post.reply_no is None or post.reply_no not in tree:
                continue
            tree[post.post_no] = post
        return OrderedDict(sorted(tree.items()))

    def get_reply_to_post(self) -> Post | None:
        """Get the post this is replying to.

        Returns:
            Parent post or None if not a reply
        """
        reply_no = self.reply_no
        if reply_no is not None:
            return Post(self._chat, post_no=reply_no).load()
        return None

    def next(self, skip_event: bool = True) -> Post | None:
        """
        Get next post in thread.

        :param bool skip_event: Skip a message if this is an event
        :return: Next post or None if no next post exists
        """
        _next = self.post_no + 1
        if _next > self._chat.posts_count:
            return None
        post = Post(self._chat, post_no=_next)
        if skip_event and post.is_event:
            return post.next()
        return post

    def prev(self, skip_event: bool = True) -> Post | None:
        """
        Get previous post in thread.

        :param bool skip_event: Skip a message if this is an event
        :return: Previous post or None if no previous post exists
        """
        _prev = self.post_no - 1
        if _prev <= 0:
            return None
        post = Post(self._chat, post_no=_prev)
        if skip_event and post.is_event:
            return post.prev()
        return post

    def who_read(self) -> dict[int, str]:
        """Get who read this post with timestamps.

        Returns:
            Dictionary mapping user IDs to read timestamps (ISO format strings)
        """
        url = f'/activity/who-read?thread_id={self._chat.id}&post_no={self.post_no}'
        response = self.client.api_get(url, response_model=WhoReadResponse)
        # Convert string user IDs to integers
        return {int(user_id): timestamp for user_id, timestamp in response.items()}

    def mark_read(self) -> bool:
        """Mark post as read.

        Returns:
            True if successfully marked as read
        """
        return self.chat.read_status(self.post_no)

    def edit(self, text: str, quote: str | None = None, reply_no: int | None = None) -> bool:
        """Edit post content.

        Args:
            text: New text content
            quote: Optional quote
            reply_no: Optional reply number

        Returns:
            True if successfully edited
        """
        url = f'/msg/post/{self._chat.id}/{self.post_no}'

        res = self.client.api_put(
            url,
            {
                'uuid': self._data.get('uuid', rand_id()),
                'text': text,
                'quote': quote,
                'reply_no': reply_no,
            },
            response_model=PostEditResponse,
        )
        if res.get('ver'):
            # Invalidate cache for this post
            if self.client._cache:
                cache_key = f'chat.{self.chat.id}.post.{self.post_no}'
                self.client._cache.delete(cache_key)

            self.load()
            return True
        return False

    def delete(self) -> bool:
        """Delete this post.

        Returns:
            True if successfully deleted
        """
        url = f'/msg/post/{self._chat.id}/{self.post_no}'
        res = self.client.api_delete(url, None, response_model=PostDeleteResponse)
        if res.get('ver'):
            # Invalidate cache for this post
            if self.client._cache:
                cache_key = f'chat.{self.chat.id}.post.{self.post_no}'
                self.client._cache.delete(cache_key)

            self.load()
            return True
        return False

    def _find_attachment(
        self,
        attachments: set[str],
        start: int,
        end: int,
    ) -> tuple[list[File], set[str]]:
        """Find attachments in post range.

        Args:
            attachments: Set of attachment UUIDs to find
            start: Start post number
            end: End post number

        Returns:
            Tuple of (found files, found UUIDs)
        """
        results: list[File] = []
        found: set[str] = set()
        for post in self.chat._lazy_posts_loader(start, end):
            if post.uuid in attachments:
                found.add(post.uuid)
            file = post.file
            if file is not None:
                results.append(file)
        return results, found

    def load_attachments(self, max_deep: int = 100, raise_if_not_found: bool = True) -> list[File]:
        """Load attachment files.

        Args:
            max_deep: Maximum depth to search for attachments
            raise_if_not_found: Whether to raise error if not all attachments found

        Returns:
            List of File objects

        Raises:
            PararamioRequestError: If not all attachments are found and raise_if_not_found is True
        """
        attachments_list = super().attachments
        if not attachments_list:
            return []
        attachments_ = set(attachments_list)
        start, end = self.post_no - len(attachments_list), self.post_no
        results, found = self._find_attachment(attachments_, start, end)
        if len(found) == len(attachments_):
            return results
        results_fb, found_fb = self._find_attachment(attachments_, end - 1 - max_deep, end - 1)
        results.extend(results_fb)
        if not raise_if_not_found or len(found_fb) + len(found) == len(attachments_):
            return results
        raise PararamioRequestError('can not find all attachments')

    @property
    def attachment_files(self) -> list[File]:
        """Get attachment files (loads them if needed)."""
        return self.load_attachments()

    @property
    def attachments(self) -> list[str]:
        """Get attachment IDs."""
        return self.meta.get('attachments', [])

    @classmethod
    def create(
        cls,
        chat: Chat,
        text: str,
        *,
        reply_no: int | None = None,
        quote: str | None = None,
        uuid: str | None = None,
        attachments: list[str] | None = None,
    ) -> Post:
        url = f'/msg/post/{chat.id}'
        data: dict[str, Any] = {
            'uuid': uuid or str(uuid4().hex),
            'text': text,
            'quote': quote,
            'reply_no': reply_no,
        }
        if attachments:
            data['attachments'] = attachments
        res = chat._client.api_post(url, data, response_model=PostCreateResponse)
        if not res:
            raise PararamioRequestError('Failed to create post')
        return cls(chat, post_no=res['post_no']).load()



"""DeferredPost API response schemas."""

from __future__ import annotations

from typing import Any, TypedDict


class DeferredPostResponse(TypedDict, total=False):
    """Schema for DeferredPost API response."""

    # IDs
    id: int
    chat_id: int
    user_id: int

    # Content
    text: str
    reply_no: int | None
    quote_range: tuple[int, int] | None

    # Timestamps (as ISO strings from API)
    time_created: str | None
    time_sending: str | None

    # Nested data
    data: dict[str, Any]


class DeferredPostsResponse(TypedDict):
    """Response for GET /msg/deferred."""

    posts: list[DeferredPostResponse]


class DeferredPostDeleteResponse(TypedDict):
    """Response for DELETE /msg/deferred/{id}."""

    result: str  # 'OK'
    deferred_post_id: int


__all__ = ['DeferredPostDeleteResponse', 'DeferredPostResponse', 'DeferredPostsResponse']



"""API endpoint constants."""

__all__ = (
    # Auth endpoints
    'AUTH_ENDPOINTS',
    'AUTH_INIT_URL',
    'AUTH_LOGIN_URL',
    'AUTH_NEXT_URL',
    'AUTH_TOTP_URL',
    'CHAT_BY_ID_URL',
    # Chat endpoints
    'CHAT_ENDPOINTS',
    'CHAT_LIST_URL',
    'CHAT_POSTS_URL',
    'CHAT_SEND_URL',
    'FILE_DELETE_URL',
    'FILE_DOWNLOAD_URL',
    # File endpoints
    'FILE_ENDPOINTS',
    'FILE_UPLOAD_URL',
    'GROUP_BY_ID_URL',
    'GROUP_JOIN_URL',
    'GROUP_LEAVE_URL',
    # Group endpoints
    'GROUP_LIST_URL',
    'POST_BY_ID_URL',
    'POST_CREATE_URL',
    'POST_DELETE_URL',
    # Post endpoints
    'POST_ENDPOINTS',
    'POST_UPDATE_URL',
    # Message endpoints
    'PRIVATE_MESSAGE_URL',
    'USER_BY_ID_URL',
    # User endpoints
    'USER_ENDPOINTS',
    'USER_PROFILE_URL',
    'USER_SEARCH_URL',
)

# Authentication endpoints
AUTH_INIT_URL = '/auth/init'
AUTH_LOGIN_URL = '/auth/login/password'
AUTH_TOTP_URL = '/auth/totp'
AUTH_NEXT_URL = '/auth/next'

# Grouped auth endpoints for convenience
AUTH_ENDPOINTS = {
    'init': AUTH_INIT_URL,
    'login': AUTH_LOGIN_URL,
    'totp': AUTH_TOTP_URL,
    'next': AUTH_NEXT_URL,
}

# User endpoints
USER_PROFILE_URL = '/user/profile'
USER_SEARCH_URL = '/user/search'
USER_BY_ID_URL = '/user/{user_id}'

# Grouped user endpoints for convenience
USER_ENDPOINTS = {
    'profile': USER_PROFILE_URL,
    'search': USER_SEARCH_URL,
    'by_id': USER_BY_ID_URL,
}

# Chat endpoints
CHAT_LIST_URL = '/chat/list'
CHAT_BY_ID_URL = '/chat/{chat_id}'
CHAT_POSTS_URL = '/chat/{chat_id}/posts'
CHAT_SEND_URL = '/chat/{chat_id}/send'

# Grouped chat endpoints for convenience
CHAT_ENDPOINTS = {
    'list': CHAT_LIST_URL,
    'by_id': CHAT_BY_ID_URL,
    'posts': CHAT_POSTS_URL,
    'send': CHAT_SEND_URL,
}

# Message endpoints
PRIVATE_MESSAGE_URL = '/msg/post/private'

# Post endpoints
POST_BY_ID_URL = '/post/{post_id}'
POST_CREATE_URL = '/post/create'
POST_UPDATE_URL = '/post/{post_id}/update'
POST_DELETE_URL = '/post/{post_id}/delete'

# Grouped post endpoints for convenience
POST_ENDPOINTS = {
    'by_id': POST_BY_ID_URL,
    'create': POST_CREATE_URL,
    'update': POST_UPDATE_URL,
    'delete': POST_DELETE_URL,
}

# Group endpoints
GROUP_LIST_URL = '/group/list'
GROUP_BY_ID_URL = '/group/{group_id}'
GROUP_JOIN_URL = '/group/{group_id}/join'
GROUP_LEAVE_URL = '/group/{group_id}/leave'

# File endpoints
FILE_UPLOAD_URL = '/upload/{perm}'
FILE_DOWNLOAD_URL = '/download/{guid}/{filename}'
FILE_DELETE_URL = '/delete/{guid}'

# Grouped file endpoints for convenience
FILE_ENDPOINTS = {
    'upload': FILE_UPLOAD_URL,
    'download': FILE_DOWNLOAD_URL,
    'delete': FILE_DELETE_URL,
}



"""Post-API response schemas."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

if TYPE_CHECKING:
    from pararamio._core import PostMetaT, TextParsedT

    from .events import PostEvent


class PostResponseItem(TypedDict, total=False):
    """Schema for a single Post item in API response."""

    # IDs from actual API response
    uuid: str
    id: int | None
    post_no: int
    user_id: int
    chat_id: int

    # Content
    text: str
    text_parsed: list[TextParsedT]

    # Reply info
    reply_no: int | None

    # Timestamps (as ISO strings from API)
    time_created: str
    time_edited: str | None

    # Metadata
    meta: PostMetaT
    event: PostEvent | None
    ver: int | None

    # Flag
    is_deleted: bool


class PostsResponse(TypedDict):
    """Schema for Posts API response wrapper."""

    posts: list[PostResponseItem]
    count: int | None  # Total count of posts


class Mention(TypedDict):
    """Mention in post."""

    id: int
    name: str
    value: str


class UserLink(Mention):
    """User link in post."""


__all__ = ['Mention', 'PostResponseItem', 'PostsResponse', 'UserLink']



"""
Logging configuration for Pararamio library.

This module provides centralized logging configuration with:
- Component-specific loggers
- Environment variable control (PARARAMIO_DEBUG)
- Performance-aware conditional logging
- Sanitization of sensitive data
"""

import logging
import os
import sys
from functools import lru_cache
from typing import Any


class LoggerManager:
    """Manages component-specific loggers with configurable levels."""

    # Component logger names
    HTTP_CLIENT = 'pararamio.http_client'
    LAZY_LOADING = 'pararamio.lazy_loading'
    BATCH_LOGIC = 'pararamio.batch_logic'
    CACHE = 'pararamio.cache'
    AUTH = 'pararamio.auth'
    SESSION = 'pararamio.session'
    API = 'pararamio.api'
    RETRY = 'pararamio.retry'
    RATE_LIMIT = 'pararamio.rate_limit'

    # Log format patterns
    DETAILED_FORMAT = (
        '%(asctime)s - %(name)s - %(levelname)s - [%(filename)s:%(lineno)d] - %(message)s'
    )
    SIMPLE_FORMAT = '%(name)s - %(levelname)s - %(message)s'
    DEBUG_FORMAT = (
        '%(asctime)s.%(msecs)03d - %(name)s - %(levelname)s - '
        '[%(filename)s:%(lineno)d in %(funcName)s()] - %(message)s'
    )

    def __init__(self) -> None:
        """Initialize the logger manager."""
        self._loggers: dict[str, logging.Logger] = {}
        self._configured = False
        self._debug_mode = self._check_debug_mode()
        self._log_level = self._get_log_level()

    @staticmethod
    def _check_debug_mode() -> bool:
        """Check if debug mode is enabled via environment variable."""
        debug_env = os.environ.get('PARARAMIO_DEBUG', '').lower()
        return debug_env in ('1', 'true', 'yes', 'on')

    def _get_log_level(self) -> int:
        """Get the appropriate log level based on environment."""
        if self._debug_mode:
            return logging.DEBUG

        level_env = os.environ.get('PARARAMIO_LOG_LEVEL', 'WARNING').upper()
        return getattr(logging, level_env, logging.WARNING)

    def _get_formatter(self) -> logging.Formatter:
        """Get the appropriate formatter based on debug mode."""
        if self._debug_mode:
            fmt = self.DEBUG_FORMAT
            datefmt = '%Y-%m-%d %H:%M:%S'
        elif self._log_level <= logging.INFO:
            fmt = self.DETAILED_FORMAT
            datefmt = '%Y-%m-%d %H:%M:%S'
        else:
            fmt = self.SIMPLE_FORMAT
            datefmt = None

        return logging.Formatter(fmt, datefmt=datefmt)

    def configure(self, force: bool = False) -> None:
        """
        Configure all loggers with appropriate handlers and levels.

        Args:
            force: Force reconfiguration even if already configured
        """
        if self._configured and not force:
            return

        # Create console handler
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(self._get_formatter())

        # Configure root pararamio logger
        root_logger = logging.getLogger('pararamio')
        root_logger.setLevel(self._log_level)
        root_logger.handlers = []  # Clear existing handlers
        root_logger.addHandler(handler)
        root_logger.propagate = False

        # Configure component loggers with specific levels if needed
        component_levels = self._get_component_levels()
        for component, level in component_levels.items():
            logger = logging.getLogger(component)
            if level is not None:
                logger.setLevel(level)

        self._configured = True

    def _get_component_levels(self) -> dict[str, int | None]:
        """Get component-specific log levels from environment variables."""
        levels = {}

        # Map of component to env variable
        env_map = {
            self.HTTP_CLIENT: 'PARARAMIO_LOG_HTTP',
            self.LAZY_LOADING: 'PARARAMIO_LOG_LAZY',
            self.BATCH_LOGIC: 'PARARAMIO_LOG_BATCH',
            self.CACHE: 'PARARAMIO_LOG_CACHE',
            self.AUTH: 'PARARAMIO_LOG_AUTH',
            self.SESSION: 'PARARAMIO_LOG_SESSION',
            self.API: 'PARARAMIO_LOG_API',
            self.RETRY: 'PARARAMIO_LOG_RETRY',
            self.RATE_LIMIT: 'PARARAMIO_LOG_RATE_LIMIT',
        }

        for component, env_var in env_map.items():
            level_str = os.environ.get(env_var)
            if level_str:
                level = getattr(logging, level_str.upper(), None)
                levels[component] = level
            else:
                levels[component] = None

        return levels

    @lru_cache(maxsize=128)
    def get_logger(self, name: str) -> logging.Logger:
        """
        Get a logger for the specified component.

        Args:
            name: Logger name (use class constants like HTTP_CLIENT)

        Returns:
            Configured logger instance
        """
        if not self._configured:
            self.configure()

        return logging.getLogger(name)

    def is_debug_enabled(self) -> bool:
        """Check if debug mode is enabled."""
        return self._debug_mode

    def set_level(self, name: str, level: int | str) -> None:
        """
        Set log level for a specific component.

        Args:
            name: Logger name
            level: Log level (int or string like 'DEBUG')
        """
        if isinstance(level, str):
            level = getattr(logging, level.upper(), logging.WARNING)

        logger = self.get_logger(name)
        logger.setLevel(level)


# Singleton instance
_manager = LoggerManager()


def get_logger(name: str) -> logging.Logger:
    """
    Get a configured logger for the specified component.

    Args:
        name: Logger name (use LoggerManager constants)

    Returns:
        Configured logger instance

    Example:
        >>> from pararamio._core.utils.logging_config import get_logger, LoggerManager
        >>> logger = get_logger(LoggerManager.HTTP_CLIENT)
        >>> logger.debug('Making API request to %s', url)
    """
    return _manager.get_logger(name)


def configure_logging(force: bool = False) -> None:
    """
    Configure the logging system.

    Args:
        force: Force reconfiguration even if already configured
    """
    _manager.configure(force=force)


def is_debug_enabled() -> bool:
    """Check if debug logging is enabled."""
    return _manager.is_debug_enabled()


def set_log_level(component: str, level: int | str) -> None:
    """
    Set log level for a specific component.

    Args:
        component: Component name (use LoggerManager constants)
        level: Log level (int or string)
    """
    _manager.set_level(component, level)


def sanitize_headers(headers: dict[str, str]) -> dict[str, str]:
    """
    Sanitize sensitive headers for logging.

    Args:
        headers: Original headers dictionary

    Returns:
        Sanitized headers with sensitive values masked
    """
    if not headers:
        return {}

    sensitive_keys = {
        'cookie',
        'authorization',
        'x-auth-token',
        'x-api-key',
        'x-csrf-token',
        'x-xsrf-token',
    }

    sanitized = {}
    for key, value in headers.items():
        key_lower = key.lower()
        if any(sensitive in key_lower for sensitive in sensitive_keys):
            # Show first 8 chars for debugging, mask the rest
            if len(value) > 8:
                sanitized[key] = f'{value[:8]}...MASKED'
            else:
                sanitized[key] = '***MASKED***'
        else:
            sanitized[key] = value

    return sanitized


def sanitize_cookies(cookies: Any) -> str:
    """
    Sanitize cookie values for logging.

    Args:
        cookies: Cookie jar or dict

    Returns:
        Sanitized representation
    """
    if not cookies:
        return 'No cookies'

    if hasattr(cookies, '__iter__'):
        count = len(list(cookies))
        return f'<{count} cookies>'
    return '<cookies present>'


def log_performance(logger: logging.Logger, operation: str, duration: float) -> None:
    """
    Log performance metrics if duration exceeds threshold.

    Args:
        logger: Logger instance
        operation: Operation name
        duration: Duration in seconds
    """
    if duration > 1.0:  # Log if operation takes more than 1 second
        logger.warning('%s took %.2f seconds', operation, duration)
    elif duration > 0.5 and logger.isEnabledFor(logging.INFO):
        logger.info('%s took %.2f seconds', operation, duration)
    elif logger.isEnabledFor(logging.DEBUG):
        logger.debug('%s took %.2f seconds', operation, duration)



"""Core Poll model without lazy loading."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Unpack

from pararamio._core.api_schemas.responses import PollOptionData, PollResponse
from pararamio._core.utils.helpers import parse_iso_datetime

from .base import CoreBaseModel

if TYPE_CHECKING:
    from datetime import datetime

    from pararamio._core._types import FormatterT

__all__ = ('CorePoll',)


# Attribute formatters for Poll
POLL_ATTR_FORMATTERS: FormatterT = {
    'time_created': parse_iso_datetime,
    'time_closed': parse_iso_datetime,
}


class CorePoll(CoreBaseModel[PollResponse]):
    """Core Poll model with common functionality."""

    _data: PollResponse
    # Poll attributes from API
    vote_uid: str
    user_id: int
    chat_id: int
    anonymous: bool
    mode: str  # 'one' or 'more'
    question: str
    time_expiration: float | None
    options: list[PollOptionData]
    total_user: int
    total_answer: int
    post_no: int | None

    # Additional attributes that a client might add
    description: str | None
    multi_choice: bool | None  # Derived from mode
    closed: bool | None
    time_created: datetime | None
    time_closed: datetime | None

    _attr_formatters: ClassVar[FormatterT] = POLL_ATTR_FORMATTERS

    def __init__(self, client: Any, **kwargs: Unpack[PollResponse]) -> None:
        """Initialize poll model with data.

        Args:
            client: Client instance (Pararamio or AsyncPararamio)
            **kwargs: Poll data
        """
        self._data = kwargs
        super().__init__(client, **kwargs)  # type: ignore[misc, call-arg]

    def __str__(self) -> str:
        return f'Poll {self._data["vote_uid"]}: {self._data["question"]}'

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CorePoll):
            return id(other) == id(self)
        return self._data['vote_uid'] == other._data['vote_uid']



"""Cookie managers for synchronous Pararamio client."""

from __future__ import annotations

import json
import logging
import os
import threading
import time
import uuid
from abc import ABC, abstractmethod
from collections.abc import Callable
from http.cookiejar import Cookie, CookieJar
from pathlib import Path
from typing import Any, TypeVar

from pararamio._core import CookieManagerBaseMixin
from pararamio._core.exceptions.base import PararamioException

T = TypeVar('T')

log = logging.getLogger(__name__)


class CookieManager(ABC):
    """Abstract cookie manager with versioning and locking support."""

    @abstractmethod
    def load_cookies(self) -> bool:
        """Load cookies from storage."""

    @abstractmethod
    def save_cookies(self) -> None:
        """Save cookies to storage."""

    @abstractmethod
    def add_cookie(self, cookie: Cookie) -> None:
        """Add or update a cookie."""

    @abstractmethod
    def get_cookie(self, domain: str, path: str, name: str) -> Cookie | None:
        """Get a specific cookie."""

    @abstractmethod
    def get_all_cookies(self) -> list[Cookie]:
        """Get all cookies."""

    @abstractmethod
    def clear_cookies(self) -> None:
        """Clear all cookies."""

    @abstractmethod
    def update_from_jar(self, cookie_jar: CookieJar) -> None:
        """Update cookies from a CookieJar."""

    @abstractmethod
    def acquire_auth_lock(self, timeout: float = 30.0) -> bool:
        """Acquire lock for the authentication process."""

    @abstractmethod
    def release_auth_lock(self) -> None:
        """Release authentication lock."""

    @abstractmethod
    def check_version(self) -> bool:
        """Check if our version matches the storage version."""

    @abstractmethod
    def refresh_if_needed(self) -> bool:
        """Reload cookies if the version changed."""

    @abstractmethod
    def handle_auth_error(self, retry_callback: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        """Handle authentication error with version check and retry."""

    def populate_jar(self, cookie_jar: CookieJar) -> None:
        """Populate a CookieJar with our cookies.

        This is a convenience method that can be implemented by subclasses
        that use CookieManagerBaseMixin.
        """
        # Default implementation - subclasses can override
        for cookie in self.get_all_cookies():
            cookie_jar.set_cookie(cookie)


class FileCookieManager(CookieManagerBaseMixin, CookieManager):
    """File-based cookie manager implementation."""

    def __init__(self, file_path: str) -> None:
        self.file_path = Path(file_path)
        self.lock_path = Path(f'{file_path}.lock')
        self.version_path = Path(f'{file_path}.version')
        self._cookies: dict[str, Cookie] = {}
        self._version: int = 0
        self._lock = threading.Lock()
        self._lock_fd: int | None = None

        # Automatically load cookies if the file exists
        if self.file_path.exists():
            try:
                self.load_cookies()
            except (OSError, ValueError):
                # Log error but don't fail initialization
                log.exception(
                    'Failed to load cookies from %s during initialization', self.file_path
                )

    @property
    def version(self) -> int:
        """Get current version."""
        return self._version

    def load_cookies(self) -> bool:
        """Load cookies from a file."""
        with self._lock:
            if not self.file_path.exists():
                return False

            try:
                with self.file_path.open(encoding='utf-8') as f:
                    data = json.load(f)

                self._load_cookies_from_dict(data)

            except (OSError, json.JSONDecodeError) as e:
                log.warning('Failed to load cookies from %s: %s', self.file_path, e)
                return False
            return True

    def save_cookies(self) -> None:
        """Save cookies to a file."""
        with self._lock:
            data = self._prepare_cookies_data()

            # Ensure directory exists
            self.file_path.parent.mkdir(parents=True, exist_ok=True)

            # Write to a temporary file first
            temp_path = Path(f'{self.file_path}.tmp')
            try:
                with temp_path.open('w', encoding='utf-8') as f:
                    json.dump(data, f, indent=2)
                # Atomic rename
                temp_path.rename(self.file_path)
            except (OSError, TypeError, ValueError):
                log.exception('Failed to save cookies to %s', self.file_path)
                if temp_path.exists():
                    temp_path.unlink()
                raise

    def add_cookie(self, cookie: Cookie) -> None:
        """Add or update a cookie."""
        with self._lock:
            key = self.make_key(cookie)
            self._cookies[key] = cookie
            self._version = self._increment_version()

    def get_cookie(self, domain: str, path: str, name: str) -> Cookie | None:
        """Get a specific cookie."""
        with self._lock:
            key = f'{domain}:{path}:{name}'
            return self._cookies.get(key)

    def get_all_cookies(self) -> list[Cookie]:
        """Get all cookies."""
        with self._lock:
            return list(self._cookies.values())

    def clear_cookies(self) -> None:
        """Clear all cookies."""
        with self._lock:
            self._cookies.clear()
            self._version = self._increment_version()

    def update_from_jar(self, cookie_jar: CookieJar) -> None:
        """Update cookies from a CookieJar."""
        with self._lock:
            for cookie in cookie_jar:
                key = self.make_key(cookie)
                self._cookies[key] = cookie
            self._version = self._increment_version()

    def acquire_auth_lock(self, timeout: float = 30.0) -> bool:
        """Acquire file lock with timeout."""
        start_time = time.time()

        while time.time() - start_time < timeout:
            try:
                # Try to create a lock file exclusively
                self._lock_fd = os.open(
                    str(self.lock_path), os.O_CREAT | os.O_EXCL | os.O_WRONLY, 0o644
                )
            except FileExistsError:
                # Lock exists, check if it's stale
                if self._is_lock_stale():
                    self._remove_stale_lock()
                    continue
                time.sleep(0.1)
                continue
            # Write PID for debugging
            os.write(self._lock_fd, str(os.getpid()).encode())
            return True

        return False

    def release_auth_lock(self) -> None:
        """Release file lock."""
        if self._lock_fd is not None:
            try:
                os.close(self._lock_fd)
                self.lock_path.unlink(missing_ok=True)
            except OSError as e:
                log.warning('Failed to release lock: %s', e)
            finally:
                self._lock_fd = None

    def check_version(self) -> bool:
        """Check if our version matches the file version."""
        current_version = self._get_file_version()
        return current_version == self._version

    def refresh_if_needed(self) -> bool:
        """Reload cookies if the version changed."""
        if not self.check_version():
            log.info('Cookie version mismatch, reloading...')
            return self.load_cookies()
        return True

    def handle_auth_error(self, retry_callback: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        """Handle authentication error with version check and retry.

        Args:
            retry_callback: Function to call for retry
            *args, **kwargs: Arguments to pass to retry_callback

        Returns:
            Result of retry_callback if successful

        Raises:
            Original exception if all retries fail
        """
        log.info('Authentication error occurred, checking cookie version...')

        # First, check if our cookies are outdated
        if not self.check_version():
            log.info('Cookie version outdated, reloading...')
            if self.load_cookies():
                log.info('Cookies reloaded, retrying with updated cookies...')
                try:
                    return retry_callback(*args, **kwargs)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    log.warning('Retry with updated cookies failed: %s', e)
            else:
                log.warning('Failed to reload cookies')

        # If still failing, acquire lock and re-authenticate
        log.info('Attempting re-authentication...')
        if self.acquire_auth_lock(timeout=30.0):
            try:
                # Clear existing cookies
                self.clear_cookies()
                self.save_cookies()

                # Call retry which should trigger re-authentication
                result = retry_callback(*args, **kwargs)

                # Save new cookies after successful auth
                self.save_cookies()
                return result

            finally:
                self.release_auth_lock()
        msg = 'Failed to acquire authentication lock for re-authentication'
        raise PararamioException(msg)

    def _increment_version(self) -> int:
        """Increment version and save to file."""
        return self._increment_file_version()

    def _is_lock_stale(self, max_age: float = 300.0) -> bool:
        """Check if the lock file is stale (older than max_age seconds)."""
        try:
            stat = self.lock_path.stat()
        except FileNotFoundError:
            return False
        age = time.time() - stat.st_mtime
        return age > max_age

    def _remove_stale_lock(self) -> None:
        """Remove stale lock file."""
        try:
            self.lock_path.unlink()
            log.info('Removed stale lock file: %s', self.lock_path)
        except FileNotFoundError:
            pass


class RedisCookieManager(CookieManagerBaseMixin, CookieManager):
    """Redis-based cookie manager implementation."""

    def __init__(self, redis_client: Any, key_prefix: str = 'pararamio:cookies') -> None:
        self.redis = redis_client
        self.key_prefix = key_prefix
        self.data_key = f'{key_prefix}:data'
        self.lock_key = f'{key_prefix}:lock'
        self.version_key = f'{key_prefix}:version'
        self._cookies: dict[str, Cookie] = {}
        self._version: int = 0
        self._lock = threading.Lock()
        self._lock_token: str | None = None

    @property
    def version(self) -> int:
        """Get current version."""
        return self._version

    def load_cookies(self) -> bool:
        """Load cookies from Redis."""
        with self._lock:
            try:
                data = self.redis.get(self.data_key)
                return self._load_cookies_from_json(data)
            except (json.JSONDecodeError, KeyError, TypeError, AttributeError) as e:
                log.warning('Failed to load cookies from Redis: %s', e)
                return False

    def save_cookies(self) -> None:
        """Save cookies to Redis."""
        with self._lock:
            data = self._prepare_cookies_data()

            try:
                self.redis.set(self.data_key, json.dumps(data))
            except (OSError, TypeError, ValueError, AttributeError):
                log.exception('Failed to save cookies to Redis')
                raise

    def add_cookie(self, cookie: Cookie) -> None:
        """Add or update a cookie."""
        with self._lock:
            key = self.make_key(cookie)
            self._cookies[key] = cookie
            self._version = self._increment_version()

    def get_cookie(self, domain: str, path: str, name: str) -> Cookie | None:
        """Get a specific cookie."""
        with self._lock:
            key = f'{domain}:{path}:{name}'
            return self._cookies.get(key)

    def get_all_cookies(self) -> list[Cookie]:
        """Get all cookies."""
        with self._lock:
            return list(self._cookies.values())

    def clear_cookies(self) -> None:
        """Clear all cookies."""
        with self._lock:
            self._cookies.clear()
            self._version = self._increment_version()

    def update_from_jar(self, cookie_jar: CookieJar) -> None:
        """Update cookies from a CookieJar."""
        with self._lock:
            for cookie in cookie_jar:
                key = self.make_key(cookie)
                self._cookies[key] = cookie
            self._version = self._increment_version()

    def acquire_auth_lock(self, timeout: float = 30.0) -> bool:
        """Acquire distributed lock using Redis."""
        self._lock_token = str(uuid.uuid4())

        start_time = time.time()
        while time.time() - start_time < timeout:
            # Try to set lock with expiration
            if self.redis.set(self.lock_key, self._lock_token, nx=True, ex=300):
                return True
            time.sleep(0.1)

        return False

    def release_auth_lock(self) -> None:
        """Release distributed lock."""
        if self._lock_token:
            # Use Lua script for atomic check-and-delete
            lua_script = """
                if redis.call("get", KEYS[1]) == ARGV[1] then
                    return redis.call("del", KEYS[1])
                else
                    return 0
                end
            """
            try:
                self.redis.eval(lua_script, 1, self.lock_key, self._lock_token)
            except (AttributeError, KeyError) as e:
                log.warning('Failed to release Redis lock: %s', e)
            finally:
                self._lock_token = None

    def check_version(self) -> bool:
        """Check if our version matches the Redis version."""
        try:
            version = self.redis.get(self.version_key)
        except (ValueError, AttributeError, TypeError, KeyError):
            return True
        current_version = int(version) if version else 0
        return current_version == self._version

    def refresh_if_needed(self) -> bool:
        """Reload cookies if the version changed."""
        if not self.check_version():
            log.info('Cookie version mismatch, reloading...')
            return self.load_cookies()
        return True

    def handle_auth_error(self, retry_callback: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        """Handle authentication error with version check and retry."""
        log.info('Authentication error occurred (Redis), checking cookie version...')

        # First, check if our cookies are outdated
        if not self.check_version():
            log.info('Cookie version outdated, reloading from Redis...')
            if self.load_cookies():
                log.info('Cookies reloaded, retrying with updated cookies...')
                try:
                    return retry_callback(*args, **kwargs)
                except Exception as e:  # pylint: disable=broad-exception-caught
                    log.warning('Retry with updated cookies failed: %s', e)
            else:
                log.warning('Failed to reload cookies from Redis')

        # If still failing, acquire distributed lock and re-authenticate
        log.info('Attempting re-authentication with distributed lock...')
        if self.acquire_auth_lock(timeout=30.0):
            try:
                # Clear existing cookies
                self.clear_cookies()
                self.save_cookies()

                # Call retry which should trigger re-authentication
                result = retry_callback(*args, **kwargs)

                # Save new cookies after successful auth
                self.save_cookies()
                return result

            finally:
                self.release_auth_lock()
        else:
            msg = 'Failed to acquire distributed lock for re-authentication'
            raise PararamioException(msg)

    def _increment_version(self) -> int:
        """Atomically increment version in Redis."""
        try:
            self._version = self.redis.incr(self.version_key)
        except (AttributeError, TypeError):
            log.exception('Failed to increment version in Redis')
            return self._version
        return self._version


class InMemoryCookieManager(CookieManagerBaseMixin, CookieManager):
    """In-memory cookie manager implementation (no persistence)."""

    def __init__(self) -> None:
        self._cookies: dict[str, Cookie] = {}
        self._version: int = 0
        self._lock = threading.Lock()

    @property
    def version(self) -> int:
        """Get current version."""
        return self._version

    def load_cookies(self) -> bool:
        """No-op for in-memory manager."""
        # Cookies are already in memory
        return bool(self._cookies)

    def save_cookies(self) -> None:
        """No-op for in-memory manager."""
        # Cookies are already in memory, nothing to save

    def add_cookie(self, cookie: Cookie) -> None:
        """Add or update a cookie."""
        with self._lock:
            key = self.make_key(cookie)
            self._cookies[key] = cookie
            self._version += 1

    def get_cookie(self, domain: str, path: str, name: str) -> Cookie | None:
        """Get a specific cookie."""
        with self._lock:
            key = f'{domain}:{path}:{name}'
            return self._cookies.get(key)

    def get_all_cookies(self) -> list[Cookie]:
        """Get all cookies."""
        with self._lock:
            return list(self._cookies.values())

    def clear_cookies(self) -> None:
        """Clear all cookies."""
        with self._lock:
            self._cookies.clear()
            self._version += 1

    def update_from_jar(self, cookie_jar: CookieJar) -> None:
        """Update cookies from a CookieJar."""
        with self._lock:
            for cookie in cookie_jar:
                key = self.make_key(cookie)
                self._cookies[key] = cookie
            self._version += 1

    def acquire_auth_lock(self, timeout: float = 30.0) -> bool:  # noqa: ARG002
        """Always returns True for in-memory manager."""
        return True

    def release_auth_lock(self) -> None:
        """No-op for in-memory manager."""

    def check_version(self) -> bool:
        """Always returns True for in-memory manager."""
        return True

    def refresh_if_needed(self) -> bool:
        """No-op for in-memory manager."""
        return True

    def handle_auth_error(self, retry_callback: Callable[..., T], *args: Any, **kwargs: Any) -> T:
        """Handle authentication error by retrying."""
        log.info('Authentication error occurred (in-memory), retrying...')

        # For in-memory, just clear cookies and retry
        self.clear_cookies()

        try:
            return retry_callback(*args, **kwargs)
        except (AttributeError, TypeError):
            log.exception('Retry failed')
            raise


__all__ = [
    'CookieManager',
    'FileCookieManager',
    'InMemoryCookieManager',
    'RedisCookieManager',
]



"""API request schemas."""

from .chat import ChatCreateRequest, ChatUpdateSettingsRequest
from .deferred_post import DeferredPostCreateRequest
from .group import GroupEditRequest
from .post import MarkReadRequest, PostCreateRequest, PostSendMessageRequest

__all__ = [
    'ChatCreateRequest',
    'ChatUpdateSettingsRequest',
    'DeferredPostCreateRequest',
    'GroupEditRequest',
    'MarkReadRequest',
    'PostCreateRequest',
    'PostSendMessageRequest',
]



"""Post API request schemas."""

from __future__ import annotations

from typing import NotRequired, TypedDict


class PostCreateRequest(TypedDict):
    """Schema for post-creation request."""

    text: str
    uuid: NotRequired[str]
    reply_no: NotRequired[int]
    quote: NotRequired[str]
    quote_range: NotRequired[tuple[int, int]]
    attachments: NotRequired[list[str]]


class PostSendMessageRequest(TypedDict):
    """Schema for simple message sending (POST /msg/post/{chat_id})."""

    uuid: str
    text: str
    reply_no: NotRequired[int]
    quote: NotRequired[str]


class MarkReadRequest(TypedDict):
    """Schema for marking chat as read."""

    read_all: NotRequired[bool]
    post_no: NotRequired[int]


__all__ = ['MarkReadRequest', 'PostCreateRequest', 'PostSendMessageRequest']



"""Deferred post-creation response schema."""

from __future__ import annotations

from typing import TypedDict


class DeferredPostCreateResponse(TypedDict):
    """Response from POST /msg/deferred endpoint."""

    deferred_post_id: str  # The ID is returned as string from API


__all__ = ['DeferredPostCreateResponse']
